"""qrem.providers.ibm module contains all necessary functions for running a characterisation/mitigation/benchmarking experiment
on IBM Quantum machines via QiskitRuntimeService or IBMServiceBackend.
"""
import os
import orjson
import math
import re
from pathlib import Path

from time import sleep
from typing import List, Union, Optional,Tuple
import numpy as np
from tqdm import tqdm
# qiskit imports
from qiskit import transpile, QuantumCircuit, QuantumRegister, ClassicalRegister, execute
from qiskit.compiler import assemble
from qiskit_ibm_runtime import IBMBackend as  IBMServiceBackend
from qiskit_ibm_runtime import QiskitRuntimeService, Session
from qiskit_ibm_provider import IBMProvider, IBMBackend

from qrem.common.providers.ibmutils import ibmutils 

# qrem imports
from qrem.common.constants import EXPERIMENT_TYPE_SYMBLOS
from qrem.common.printer import qprint, errprint, warprint
from qrem.common import convert, io
from qrem.common.utils import get_full_object_size
from qrem.circuit_collection import CircuitCollection
from qrem.experiment_results import ExperimentResults


QISKIT_IBM_TOKEN = os.getenv('QISKIT_IBM_TOKEN')
QISKIT_IBM_CHANNEL = os.getenv('QISKIT_IBM_CHANNEL')


#Hardcoded limit for job  input size in megabytes
JOB_PAYLOD_SIZE_LIMIT = 15 #Mb


#-----------------------
# PART 0: Connect to the IBM Quantum specific backend using your token 
#-----------------------
#TODO (PP) in theory you can check which backend is the best and take the best one from the backends for a fiven processor list. Would be nice to implement
# service.least_busy(operational=True, min_num_qubits=5)
# service.backends(simulator=False, operational=True, min_num_qubits=5)
#which connection method we will use

#Optional instance: 'ibm-q-psnc/internal/reservations'
def connect(name:str = "", 
            channel: Optional[str] = QISKIT_IBM_CHANNEL, 
            method:str = "RUNTIME", #available are "RUNTIME", "PROVIDER", "RUNTIME SESSIONS", "JOB EXECUTE"
            instance: Optional[str] = 'ibm-q/open/main', 
            verbose_log = False) -> Tuple[Union[IBMServiceBackend,IBMBackend],Optional[QiskitRuntimeService],Optional[IBMProvider]]:
    """
    Initializes connection to the backend based on the API_TOKEN defined in the environment variables.
    Depending on the <method> parameter will start either QiskitRuntimeService object or IBMProvider object

    Parameters
    ----------
    channel: str, default='ibm_quantum' 
        specifies which channel to use for connection (ibm_quantum or ibm_cloud)
    method: str, default='RUNTIME' 
        specifies connection method. Available are "RUNTIME", "PROVIDER", "RUNTIME SESSIONS", "JOB EXECUTE"
    instance: str, default='ibm-q/open/main' 
        specifies instance for connection, example values are: 'ibm-q/open/main' or 'ibm-q-psnc/internal/reservations' 
    verbose_log: bool, default=False
        turn on verbose logging for more printouts with info
        
    Returns
    -------
    backend
        Instance of backend object returned by QiskitRuntimeService (IBMServiceBackend) or IBMProvider (IBMBackend)
    service
        Connected Instance of QiskitRuntimeService. Null if connecting via IBMProvider
    provider
        Connected Instance of IBMProvider. Null if connecting via QiskitRuntimeService
    """   

    #[1] Connect and save provider/service objects
    service = None
    provider = None
    if "RUNTIME" in method:
        service = _connect_via_runtime(backend_name= name, channel = channel, instance = instance, verbose_log = verbose_log)
        backend = service.get_backend(name)
    elif "DEBUG" in method:
        service = None
        backend = "DEBUG"
        
    else:
        provider = _connect_via_provider(backend_name= name, channel = channel, instance = instance, verbose_log = verbose_log)
        backend = provider.get_backend(name)    

    if backend != "DEBUG":
        #[2] Check status of the backened
        status = backend.status()
        is_operational = status.operational
        is_internal = status.status_msg
        jobs_in_queue = status.pending_jobs
        max_circuits = backend.max_circuits
    else:
        is_operational = "debug"
        is_internal = "mode"
        jobs_in_queue = 0
        max_circuits = 127

    if verbose_log:
        
        qprint("----------")
        qprint(f"The backend {name} status:", f"{is_operational} : {is_internal}")
        qprint(f"- jobs in queue:", f"{jobs_in_queue}")
        qprint(f"- official circuits limit:", f"{max_circuits}")
              

    #use this first to connect to a backend
    if not is_operational or is_internal == "internal":
        errprint(f"ERROR: The backend {name} is currently not available. ", f"Confirm on https://quantum-computing.ibm.com/services/resources ")
    
    return backend, service, provider

def _connect_via_runtime(backend_name:str = "", channel: Optional[str] = 'ibm_quantum', instance: Optional[str] = 'ibm-q/open/main', verbose_log = False):
    """
    Helper function for qrem.common.providers.ibm.connect(...) to connect via QiskitRuntimeService
    """
    if instance == '':
        service = QiskitRuntimeService(channel = channel, token=QISKIT_IBM_TOKEN)
    else:
        service = QiskitRuntimeService(channel = channel, token=QISKIT_IBM_TOKEN, instance = instance)

    return service
  
def _connect_via_provider(backend_name:str = "", channel: Optional[str] = 'ibm_quantum', instance: Optional[str] = 'ibm-q/open/main', verbose_log = False):
    """
    Helper function for qrem.common.providers.ibm.connect(...) to connect via IBMProvider
    """
    if instance == '':
        provider = IBMProvider(token=QISKIT_IBM_TOKEN)
    else:
        provider = IBMProvider(token=QISKIT_IBM_TOKEN, instance = instance)

    return provider


#-----------------------
# PART 1: Get backend properties (most importantly - good qubits)
#-----------------------
def get_valid_qubits_properties(  backend: Union[IBMBackend,IBMServiceBackend], 
                            threshold: Optional[float] = 0,
                            verbose_log = False ):
    """
    get_valid_qubits_properties queries the backend object to provide information about which qubits are valid
    (not flagged faulty by IBM and below given sx gate error treshold ). Returns an elaborate dictionary with 

    Parameters
    ----------
    backend: Union[IBMBackend,IBMServiceBackend]
        Instance of backend object returned during connection by QiskitRuntimeService (IBMServiceBackend) or IBMProvider (IBMBackend)
    threshold: floatm default = 0
        gate error treshold. If 0 (default value), will not discard qubits based on gate error tresholding
    verbose_log: bool, default=False
        turn on verbose logging for more printouts with info
        
    Returns
    -------
    dict
        Dictionary describing valid qubits for experiment. 
        {"good_qubits_indices": list of valid qubit indices,
        "number_of_good_qubits": count of valid qubits,
        "original_quantum_register_size": orginal number of available qubits,
        "faulty_qubits": list of qubits indices marked faulty by IBM at the time of connection,
        "non_operational_qubits": list of qubit indices above sx gate error treshold,
        "readout_errors": list of returned readout errors on each of the qubit}
    """       

    if backend == "DEBUG":
        #  DEBUG MODE - backend is null
        qprint("Running in a DEBUG mode (backend is Null)")
        return {"good_qubits_indices": list(range(0,127)),
                "number_of_good_qubits": 127,
                "original_quantum_register_size": 127,
                
                "faulty_qubits": [],
                "operational_qubits": list(range(0,127)),
                "readout_errors": list (0.001*np.random.sample(127))}
    elif backend == None:
        errprint("Backend is null - something went wrong with connection.")

    #TODO (PP) test on a fake propetries object

    properties = backend.properties()
    # - quantum register size
    quantum_register_size = backend.configuration().n_qubits

    # [1] get faulty qubits -> Important for some quantum machines like Osprey
    faulty_qubits = [q for q in range(quantum_register_size) if q in properties.faulty_qubits()]

    # [2] get qbits above treshold
    if(threshold!= None and threshold > 0):
    # [2.1] for each qubit in backend get single qubit gate errors (of gate sqrt(X), which is treated as a model gate):
        gate_errors_list = []
        for qubit in range(quantum_register_size):
            gate_errors_list.append(properties.gate_error('sx', qubit))

        qbits_above_error_treshold = [idx for idx, val in enumerate(gate_errors_list) if val > threshold]
    else:
        qbits_above_error_treshold = []

    # SHOULD WE ALSO CHECK FOR OPERATIONAL QUBITS?  FOR NOW SKIPPED
    operational_qubits = [q for q in range(quantum_register_size) if properties.is_qubit_operational(q)]
    
    # ANYTHING WE SHOULD DO ABOUT READOUT ERRORS? FOR NOW SKIPPED
    readout_errors = [properties.readout_error(q) for q in range(quantum_register_size)]

    # [3] get correct qubits
    all_bad_qubits = faulty_qubits + qbits_above_error_treshold #we don't need to make a correct set union here, but it would be more proper
    good_qubits_indices =  [q for q in range(quantum_register_size) if q not in all_bad_qubits]
    # [3.1] Renmove double entries and sort indices to ensure, that results can easily be interpreted. (NOT NECESSARY HERE MOST LIKELY)
    good_qubits_indices = list(set(good_qubits_indices)) 

    # [3.2] Important - indicies should be always sorted
    good_qubits_indices = sorted(good_qubits_indices) 

    number_of_good_qubits = len(good_qubits_indices)

    return {"good_qubits_indices": good_qubits_indices,
            "number_of_good_qubits": number_of_good_qubits,
            "original_quantum_register_size": quantum_register_size,
            
            "faulty_qubits": faulty_qubits,
            "operational_qubits": operational_qubits,
            "readout_errors": readout_errors}
    pass



#-----------------------
# PART 2: Translate circuits to IBM / QISKIT FORMAT
#-----------------------
def translate_circuits_to_qiskit_format(circuit_collection:CircuitCollection) -> List[QuantumCircuit]:
    """
    Translates circuits defined as ndarray of unit8s with QREM internal circuit labelling (labels defined in qrem.common.constants.EXPERIMENT_TYPE_SYMBLOS)
    into a list of Qiskit Circuits (not yet transpiled)

    Parameters
    ----------
    circuit_collection: CircuitCollection
        Instance of CircuitCollection object, needs to be already initialized with chosen circuits, proper valid qubit_indices and all other fields.
        
    Returns
    -------
    qiskit_circuits: QuantumCircuit
        Lsit of QuantumCircuits, with barrier between each of the qubit (read more on qiskit docs about QuantumCircuits)
    """     

    # TODO (PP) sort out repeating circuits already in CircuitCollection
    # prerequisite: no repeating circuits in the accepted collection
    qiskit_circuits = []

    quantum_register_size = max(circuit_collection.qubit_indices) + 1
    classical_register_size = len(circuit_collection.qubit_indices)

    for id in range(0,len(circuit_collection.circuits)):
        #[1] Make a label string from circuit labels
        #TODO (PP) won't work for machines with 1 million qubits
        base_label = np.array2string(circuit_collection.circuits[id], precision=0, separator='', max_line_width = 1000000)[1:-1]
        label = circuit_collection.experiment_name + "___" + base_label + "___" +str(id)


        #[2]  Make an empty circuit
        q_reg = QuantumRegister(quantum_register_size)
        c_reg = ClassicalRegister(classical_register_size)
        qiskit_circuit = QuantumCircuit(q_reg, c_reg, name=label)
    
        #[3]  Fill in according to experyment type:
        if(circuit_collection.experiment_type.lower() == "ddot"):
            ibmutils.circuits_translator_DDOT(  circuit=circuit_collection.circuits[id], 
                                        qiskit_circuit = qiskit_circuit,
                                        valid_qubit_indices=circuit_collection.qubit_indices) 
        elif(circuit_collection.experiment_type.lower() == "qdot"):
            ibmutils.ccircuits_translator_QDOT(  circuit=circuit_collection.circuits[id],  
                                        qiskit_circuit = qiskit_circuit,
                                        valid_qubit_indices=circuit_collection.qubit_indices) 
        
        else:
            errprint(f"ERROR: not supported experiment type:",f"{circuit_collection.experiment_type}")
        #[4] Barrier between all qubits (no optimizartion allowed)
        qiskit_circuit.barrier()

        ibmutils.add_measurements( qiskit_circuit = qiskit_circuit,
                            q_reg=q_reg,
                            c_reg=c_reg,
                            valid_qubit_indices=circuit_collection.qubit_indices)


        #[5] Append to list of all circuits
        qiskit_circuits.append(qiskit_circuit)

    return qiskit_circuits


def _add_measurements(
        
    qiskit_circuit:QuantumCircuit,
    q_reg:QuantumRegister,
    c_reg:ClassicalRegister,
    valid_qubit_indices:List[int]):
    """
    Helper function that adds a measurement for every valid qubit defined in valid_qubit_indices list on existing qiskit_circuit (QuantumCircuit).
    """  
    # - we need to measure for every available qubit (number of qubits = classical_register_size = len(valid_qubit_indices)),
    # from a valid qubit with an index taken from a sorted list of valid_qubit_indices
    # - we only need as many qubits as max defined value of valid indices in our circuit definition (q_reg.size)

    for qubit_index in range(c_reg.size):
        which_qubit = q_reg[valid_qubit_indices[qubit_index]]
        which_classical_output = c_reg[qubit_index]
        qiskit_circuit.measure(which_qubit, which_classical_output)  

def _circuits_translator_DDOT( circuit:np.ndarray, 
                                qiskit_circuit:QuantumCircuit,
                                valid_qubit_indices:list) -> QuantumCircuit:
    """
    Helper method that creates DDOT circuits in format native to QISKIT and applies to existing (empty) qiskit_circuit,
    based on the passed circuits and valid_qubit_indices, specifying what gates should be applied to what qubits.
    """

    
    for qubit_index in range(circuit.size):

        qubit_id = valid_qubit_indices[qubit_index]
        if circuit[qubit_index] == 0 :
            qiskit_circuit.id(qubit_id)
        elif circuit[qubit_index] == 1:
            qiskit_circuit.x(qubit_id)
        else:
            raise ValueError('Wrong circuit label: ', circuit)

    
    return qiskit_circuit

def _circuits_translator_QDOT( circuit:np.ndarray, 
                                qiskit_circuit:QuantumCircuit,
                                valid_qubit_indices:list) -> QuantumCircuit:
    """
    Helper method that creates QDOT circuits in format native to QISKIT and applies to existing (empty) qiskit_circuit,
    based on the passed circuits and valid_qubit_indices, specifying what gates should be applied to what qubits.
    """
    # Opisac to lepiej i dodac ze to konkretnie dla Pauliego
        
    
    for qubit_index in range(circuit.size):
       
        qubit_id = valid_qubit_indices[qubit_index]

        if circuit[qubit_index] == 0:
            qiskit_circuit.id(qubit_id)
        elif circuit[qubit_index] == 1:
            qiskit_circuit.x(qubit_id)
        elif circuit[qubit_index] == 2:
            qiskit_circuit.h(qubit_id)
        elif circuit[qubit_index] == 3:
            qiskit_circuit.x(qubit_id)
            qiskit_circuit.h(qubit_id)
        elif circuit[qubit_index] == 4:
            qiskit_circuit.h(qubit_id)
            qiskit_circuit.s(qubit_id)
        elif circuit[qubit_index] == 5:
            qiskit_circuit.x(qubit_id)
            qiskit_circuit.h(qubit_id)
            qiskit_circuit.s(qubit_id)
        else:
            raise ValueError(
                f"Wrong circuit label: '{circuit[qubit_index]} from circuit: '{circuit}'")

    return qiskit_circuit

#-----------------------
# PART 2: Run circuits
#-----------------------
def execute_circuits(qiskit_circuits:List[QuantumCircuit],
                     method:str = "RUNTIME",
                     job_tags:List[str] = ["QREM_JOB",],
                     instance:str = 'ibm-q/open/main',
                     service:QiskitRuntimeService = None,
                     backend:Union[IBMServiceBackend,IBMBackend] = None,
                     number_of_repetitions:int = 1024,
                     log_level:str ='WARNING',
                     verbose_log:bool = False):
    """
    Executes circuits prepared in QISKIT format (qiskit_circuits) on the backend provided by a qrem.common.providers.ibm.connection(...) function.

    Parameters
    ----------
    qiskit_circuits: List[QuantumCircuit]
        List of QuantumCircuits, not yet transpiled, to be executed on chosen IBM machine
    method: str, default = "RUNTIME"
        Connection method, determined by the execution of qrem.common.providers.ibm.connection(...) (should be same as there)
    job_tags: List[str] default=["QREM_JOB"]
        List of tags to be applied to each job
    service: QiskitRuntimeService, default=None
        Connected Instance of QiskitRuntimeService. Can be None if connecting via IBMProvider
    backend: Union[IBMServiceBackend,IBMBackend]
        Provide connected Backend here
    number_of_repetitions:int, default=1024
        Number of repetitions for the circuit collection (persistently named number of shots by providers)
    log_level: str, default = "WARNING"
        Logging level to log job execution on machine. The valid log levels are: DEBUG, INFO, WARNING, ERROR, and CRITICAL. The default level is WARNING.
        look https://qiskit.org/ecosystem/ibm-runtime/stubs/qiskit_ibm_runtime.RuntimeOptions.html#qiskit_ibm_runtime.RuntimeOptions for more information
        
    Returns
    -------
    job_id_list: List[str]
        Lsit of submitted job IDs
    """         
    
    
    #TODO (PP) check for null Service, provider and backend type - for DEBUG and other purposes

    # The valid log levels are: DEBUG, INFO, WARNING, ERROR, and CRITICAL. The default level is WARNING.
    # translate to IBM - qiskit Circuits
    if(number_of_repetitions > backend.max_shots):
        warprint(f"WARNING: Number of repetitions ({number_of_repetitions}) is larger than max number of shots ({backend.max_shots}) defined on chosen backend {backend.name}, defaulting to: ", f"({backend.max_shots})")
    # transpile all circuits
    transpiled_circuits = ibm_translate.transpile_circuits(qiskit_circuits,backend)

    #[X] Caculate size of a given batch:
    qobject = assemble (experiments = transpiled_circuits,shots = number_of_repetitions)
    full_payload_size = get_full_object_size(qobject)/1000000.0

    number_of_circuits = len(transpiled_circuits)
    batch_division = math.ceil(full_payload_size/JOB_PAYLOD_SIZE_LIMIT)
    batch_size = min(backend.max_circuits-1, math.floor(number_of_circuits/batch_division))
    if batch_size > backend.max_circuits:
        warprint(f"WARNING: Calculated batch circuit count based on payload size limit ({batch_size}) is larger than machine max circuits count. Defaulting to max circuits count -1:", f"{backend.max_circuits-1}")
    #[X] divide into batches based on the calculated quantities
    batches = [transpiled_circuits[i:i + batch_size] for i in range(0, number_of_circuits, batch_size)] 

    job_id_list = []
    # choose specific method (ibm_provider vs ibm_runtime?)
    if method == "RUNTIME_SESSIONS":
        if verbose_log:
            qprint(f"Now sending  prepared jobs to the IBM backend {backend.name} via Runtime with sessions:", f"{len(qiskit_circuits)} jobs repeated {number_of_repetitions} times")
            qprint(f"- logging level:", log_level)
            qprint(f"- IBM connection instance:", instance)
            qprint(f"- memory:", "True")
            qprint(f"- meas_level:", 2)
            qprint(f"- number of repetitions:", number_of_repetitions)
            qprint(f"- job tags:", job_tags)
            warprint(f"-------------------")

        runtime_options = {
            'backend': backend.name,
            'job_tags': job_tags,
            #'optimization_level': 0, #no circuit optimisation
            'log_level': log_level, 
            'instance': instance
        }

        #divide 
        with Session(service=service, backend=backend) as session:
            job_list = []
            count = 0
            tdqm_batches = tqdm(batches)
            for batch in tdqm_batches:
                # more in https://quantum-computing.ibm.com/services/programs/prototypes?program=circuit-runner
                runtime_inputs = {
                    'circuits': batch,
                    'shots': number_of_repetitions,  # integer
                    'memory': True,  # integer#per-shot measurement bitstrings
                    'meas_level': 2, # The measurement output 2 is the discriminated measurement counts. Level 1 is the IQ measurement kernel values.
                }

                job = service.run(
                    program_id='circuit-runner',
                    options=runtime_options,
                    inputs=runtime_inputs,
                )
                sleep(0.5)
                qprint(f"Sent job ({count}/{len(batches)}) ")
                job_list.append(job)
                job_abbrv = str(job.job_id())
                job_abbrv =  (job_abbrv[6] + '..') if len(job_abbrv) > 8  else job_abbrv
                tdqm_batches.set_description(f"Sent job ({job_abbrv}):")

                # XXXXXXXXXXXXXXXXXXXXXXXXX

                # job.result() is blocking, so this job is now finished and the session can be safely closed.
                # Close the session only if all jobs are finished, and you don't need to run more in the session
            sleep(0.5) #not sure if necessarry?
            for job in job_list:
                job_id_list.append(job.job_id())
                # Do not close here, the job might not be completed!
                result = job.result()
                #TODO (PP) UNCERTAIN: shall we already save this result?
            session.close()

    elif method == "RUNTIME":
        if verbose_log:
            qprint(f"Sending  prepared jobs to the IBM backend {backend.name} via Runtime without sessions:", f"{len(qiskit_circuits)} jobs repeated {number_of_repetitions} times")
            qprint(f"- logging level:", log_level)
            qprint(f"- IBM connection instance:", instance)
            qprint(f"- memory:", "True")
            qprint(f"- meas_level:", 2)
            qprint(f"-------------------")
            qprint(f"- number of repetitions:", number_of_repetitions)
            qprint(f"- job tags:", job_tags)

        runtime_options = {
            'backend_name': backend.name,
            'job_tags': job_tags,
            'log_level': log_level, #no mitigation
            'instance': instance
        }
        
        for batch in batches:
            runtime_inputs = {
                'circuits': batch,
                'shots': number_of_repetitions,  # integer
            }
            job = service.run(
                program_id='circuit-runner',
                options=runtime_options,
                inputs=runtime_inputs,
            )
            sleep(0.5) #not sure if necessarry?
            job_id_list.append(job.job_id())

    elif method == "JOB_EXECUTE":
        if verbose_log:
            qprint(f"Sending  prepared jobs to the IBM backend {backend.name} via Qiskit Job.execute:", f"{len(qiskit_circuits)} jobs repeated {number_of_repetitions} times")
            qprint(f"- memory:", "True")
            qprint(f"- optimization_level:", 0)
            qprint(f"-------------------")
            qprint(f"- number of repetitions:", number_of_repetitions)
            qprint(f"- job tags:", job_tags)   

        for batch in batches:
            job = execute(
                            experiments=batch, 
                            backend=backend, 
                            shots=number_of_repetitions, 
                            optimization_level = 0, 
                            memory=True)
            sleep(0.5) #not sure if necessarry?
            job_id_list.append(job.job_id())
            job.update_tags(job_tags)
            sleep(0.1)
        pass

    elif method == "PROVIDER":
        if verbose_log:
            qprint(f"Sending  prepared jobs to the IBM backend {backend.name} via IBMProvider Bakend.run:", f"{len(qiskit_circuits)} jobs repeated {number_of_repetitions} times")
            qprint(f"-------------------")
            qprint(f"- number of repetitions:", number_of_repetitions)
            qprint(f"- job tags:", job_tags)   

        for batch in batches:
            job = backend.run(circuits=batch, shots = number_of_repetitions)
            sleep(0.5) #not sure if necessarry?
            job_id_list.append(job.job_id())
            job.update_tags(job_tags)
            sleep(0.1)
    pass

    return job_id_list

#-----------------------
# PART 3: Not used yet
#-----------------------
def retrieve_results(provider, 
                    job_IDs:List[str],
                    original_circuits: Union[CircuitCollection,bool]= False,
                    original_circuits_path: Union[bool,str,Path] = False,
                    save_experiment_results: Union[bool,str,Path]=False,
                    backup_original: Union[bool,str,Path]=False,
                    overwrite = False):   
    

    #[1] Download original results from provider
    original_results, circuit_labels = ibmutils.download_results(provider = provider, 
                                                                 job_IDs = job_IDs,
                                                                 backup =  backup_original,
                                                                 overwrite = overwrite);
    
    #[2] Retrieve, prepare circuits collection; look for discrepancies
    if(original_circuits_path):
        #this will compare circuits and provide original circuits if they are not provided in original_circuits
        original_circuits = ibmutils.compare_circuits(circuit_labels, original_circuits_path, original_circuits );

    #[3] Retrieve, prepare circuits collection; look for discrepancies
    if not original_circuits:
        warprint("METADATA MISSING. No original circuits provided. ExperimentResults will miss metadata")
        original_circuits = CircuitCollection("Unknown")
        original_circuits.circuits = [ convert.bitstring_to_ndarray(label) for label in circuit_labels ]
        original_circuits.circuits_labels = circuit_labels
    experiment_results = ExperimentResults()
    experiment_results.from_ibm(qrem_circuits = original_circuits, original_results = original_results, circuit_labels = circuit_labels)
    
    
    export_path = io.prepare_outfile(outpath = save_experiment_results, overwrite = overwrite , default_filename= 'experiment_results.pkl')
    experiment_results.export_json(json_export_path = export_path)

    return 
    




# test:
if __name__ == "__main__":
    pass
    # from datetime import datetime

    # from qrem.types import CircuitCollection
    # from qrem.common.experiment.tomography import compute_number_of_circuits, create_circuits

    # experiment_type = 'DDOT'
    # device_name = 'ibmq_kolkata'
    # provider_instance = ''  # TODO: input something

    # provider = IBMProvider(instance=provider_instance)
    # backend = provider.get_backend(device_name)

    # [0] get from provider information on which qubits have gate error belowe threshold:
    # qubit_indices = get_qubits_below_threshold(backend)
    # no_qubits = len(qubit_indices)

    # # [1] create the circuits:
    # no_circuits = compute_number_of_circuits(experiment_type=experiment_type, number_of_qubits=no_qubits)
    # circuit_collection = create_circuits(experiment_type, no_qubits, no_circuits)
    # #TODO: [1] should result in a CircuitCollection
    # #TODO: add checking completeness
    #
    # # PLACEHOLDER instead of [1]:
    # circuit_collection = CircuitCollection('test_name')
    # print(circuit_collection.get_dict_format())
    # dictionary_to_load = {'experiment_type': 'qdot',
    #                       'circuits_list': [[0, 5, 3], [1, 2, 0], [3, 1, 4]],
    #                       'qubit_indices': [0, 2, 5],
    #                       'gate_error_threshold': 0.005,
    #                       'no_shots': 1,
    #                       'datetime_created_utc': datetime.utcnow(),
    #                       'author': 'tester',
    #                       'notes': 'some string note'}
    # circuit_collection.load_from_dict(dictionary_to_load)

    # [2] execute the circuits:
    # job_IDs = execute_circuits(device_name, circuit_collection)

    # [3] download results:
    # job_IDs = ["",]#TODO: input something
    # results_qiskit = get_results(provider, job_IDs)
    # results_qrem =




#----------------------------------------------------
#----------------------------------------------------
#---IBM circuit_runner runtime inputs (runtime_inputs)
# TODO (PP) (MO) talk options for init_qubits,  memory,  use_measure_esp
#----------------------------------------------------
#----------------------------------------------------
	# # A circuit or a listof QuantumCircuits or a list
	# # of QASM 2.0 strings orna Dict of QasmQobj/PulseQobj.
	# 'circuits': None, # [object,array] (required)

	# # Number of repetitions of each circuit, for sampling. Default: 1024.
	# 'shots': 4000, # integer

	# # User input that will bemattached to the job and will be copied to the corresponding result header.
	# 'header': None, # object

	# # Whether to reset the qubits to the ground state for each shot.
	# 'init_qubits': True, # boolean

	# # The measurement output level. Level 2 is the discriminated measurement counts. Level 1 is the IQ measurement kernel values.
	# 'meas_level': 2, # integer

	# Whether to return per-shot measurement bitstrings.
	# 'memory': False, # boolean
    
	# Whether to use excited state
	# promoted (ESP) readout for measurements,
	# which are the terminal instruction
	# to a qubit. ESP readout
	# can offer higher fidelity than
	# standard measurement sequences.
	# 'use_measure_esp': None # boolean

	# # Delay between programs in seconds.
	# 'rep_delay': None, # number

	# # List of measurement LO frequencies in Hz. Overridden by schedule_los if specified.
	# 'meas_lo_freq': None, # array

	# Type of measurement data to return. Only applicable for meas_level=1.
	# If 'single' is specified, per-shot information is returned. If 'avg'
	# is specified, average measurement output is returned.
	# 'meas_return': None, # string


	# # List of job level qubit drive LO frequencies in Hz. Overridden by schedule_los if specified.
	# 'qubit_lo_freq': None, # array

	# Experiment level LO frequency configurations for qubit drive and measurement channels, in Hz.
	# 'schedule_los': None, # array




import os
import pickle
from braket.aws import AwsDevice

from braket.circuits import Circuit

def append_new_line(file_name, text_to_append):
    """
    Helper function to store job ids.
    Append given text as a new line at the end of file
    """
    # Open the file in append & read mode ('a+')
    with open(file_name, "a+") as file_object:
        # Move read cursor to the start of file.
        file_object.seek(0)
        # If file is not empty then append '\n'
        data = file_object.read(100)
        if len(data) > 0:
            file_object.write("\n")
        # Append text at the end of file
        file_object.write(text_to_append)

#-----------------------
# PART 2: Translate circuits to IBM / QISKIT FORMAT
#-----------------------
def _circuits_translator(eigenstate_index: int,
                            quantum_circuit: Circuit,
                            qubit_index: int):
    """
    Helper method that creates both DDOT and QDOT circuits in format native to BRAKET and applies to existing (empty) braket_circuit,
     by applying Pauli eigenstates (formerly _apply_pauli_eigenstate).
    """
    # _pauli_labels = ['z+', 'z-', 'x+', 'x-', 'y+', 'y-']

    # Z+
    if eigenstate_index == 0:
        quantum_circuit = quantum_circuit.rz(qubit_index, 0)

    # Z-
    elif eigenstate_index == 1:
        # quantum_circuit = quantum_circuit.rx(qubit_index, np.pi)
        quantum_circuit = quantum_circuit.rz(qubit_index, -0.9199372448290238)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi)
        quantum_circuit = quantum_circuit.rx(qubit_index, - np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, 2.2216554087607694)

    # X+
    elif eigenstate_index == 2:
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)

    # X-
    elif eigenstate_index == 3:
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, -np.pi / 2)

    # Y+
    elif eigenstate_index == 4:
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi)

    # Y-
    elif eigenstate_index == 5:
        quantum_circuit = quantum_circuit.rz(qubit_index, np.pi / 2)
        quantum_circuit = quantum_circuit.rx(qubit_index, np.pi / 2)

    else:
        raise ValueError(f"Incorrect eigenstate index: '{eigenstate_index}'!")

    return quantum_circuit

def run():
    """
    Entrypoint function to run prepared experiment.
    See https://amazon-braket-sdk-python.readthedocs.io/en/stable/_apidoc/braket.aws.aws_quantum_job.html
    Write in create  AwsQuantumJob: entry_point="aws_braket_runner:run"
    Should run until all job/tasks are completed
    """
    #[1]Entry definityions
    device = AwsDevice(os.environ["AMZN_BRAKET_DEVICE_ARN"])
    input_dir = os.environ["AMZN_BRAKET_INPUT_DIR"]
    output_dir = os.environ["AMZN_BRAKET_JOB_RESULTS_DIR"]

    circuits_file = f'{input_dir}/braket_circuits_list/braket_circuits_list.pkl'
    circuits_symbolic_file =  f'{input_dir}/qrem_circuits_list/qrem_circuits_list.txt'
    metadata_file = f'{input_dir}/metadata/metadata.pkl'
    good_qubits_indices_file = f'{input_dir}/good_qubits_indices/good_qubits_indices.txt'
    task_arns_file = f'{output_dir}/task_arns.txt'


    #[0] Prepare valid qubit indices list
    #------------------------------------
    good_qubits_indices = []
    with open(good_qubits_indices_file, 'r') as fp:
        for line in fp:
            # remove linebreak from a current name
            # linebreak is the last character of each line
            x = line[:-1]
            good_qubits_indices.append(int(x))
    
    #[1] Load pickled braket circuits and metadata
    #---------------------------------------------
    braket_circuits =[]
    circuits_symbolic = []
    results = []  #ADD
    task_arn_list = []  #ADD
    metadata = {}
    pickle_load = False

    with open(metadata_file, 'rb') as fm:
        metadata = pickle.load(fm) # deserialize using load()
        

    with open(circuits_symbolic_file, 'r') as fp:
        for line in fp:
            # remove linebreak from a current name
            # linebreak is the last character of each line
            x = line[:-1]
            new_circuit = []
            for label in x:
                new_circuit.append(int(label))
            circuits_symbolic.append(new_circuit)


    if os.path.isfile(circuits_file):
        with open(circuits_file, 'rb') as f:
            braket_circuits = pickle.load(f) # deserialize using load()
    #[1.1] If pickled circuits were not provided:
    #---------------------------------------------
    else:
        circuits_number = len(circuits_symbolic)
        for circuit_idx in range(circuits_number):
            aws_circuit = Circuit()
            for idx, label in enumerate(circuits_symbolic[circuit_idx]):
                aws_circuit = _circuits_translator(eigenstate_index=label, quantum_circuit=aws_circuit,
                                                    qubit_index=good_qubits_indices[idx])
            aws_circuit = Circuit().add_verbatim_box(aws_circuit)
            braket_circuits.append(aws_circuit)


    # [2] prepare tags if present in metadata
    #------------------------------------------------------------------------------------------------------------
    number_of_shots = metadata["number_of_repetitions"]
    MAX_RETRIES = metadata.get("max_task_retries", 3)  
    tags = metadata.get("tags", {}) 
    
    # [3] run tasks consecutively. According to aws, this ensures order of exxecution. We also save circuit number as a hash
    #------------------------------------------------------------------------------------------------------------
    #TODO czy circuit to jest reprezentacja binarna czy obiekt braketu?
    tasks = [device.run(circuit, 
                        shots=number_of_shots, 
                        disable_qubit_rewiring=True, 
                        tags={**tags, "circuit_id": idc, "circuit": circuits_symbolic[idc]}) 
                        for (idc, circuit) in enumerate(braket_circuits)]
     
    # now tasks will be saved on s3. However, we can aggregate all results in a single results file in output folder - point [5] below
    # [4] retry failed tasks
    #------------------------------------------------------------------------------------------------------------
    
    retries = 0

    results = [task.result() for task in tasks]
    while retries < MAX_RETRIES and not all(results): 
        for i, result in enumerate(results):
            if not result:
                print(f"Task {tasks[i].arn} had failed. Retrying")
                tasks[i] =device.run(braket_circuits[i], 
                                    shots=number_of_shots, 
                                    disable_qubit_rewiring=True, 
                                    tags={**tags, "circuit_id": i, "circuit": circuits_symbolic[i]}) 
        results = [task.result() for task in tasks]
        retries += 1      

    failed_tasks = [(i, task.arn) for (i, task) in enumerate(tasks) if not results[i]]
    
    # [5] Save job results and arns of failed tasks
    #------------------------------------------------------------------------------------------------------------
    
    task_arn_list = [task.arn for task in tasks]
    
    #TODO unikalność circuits - czy jest unikalne?
    measurements = {str(circuit):result.measurement_counts if result else None for (circuit,result) in zip(circuits_symbolic,results) }
    save_job_result({ "measurement_counts": measurements, 
                      "task_arn_list": task_arn_list, 
                      "failed_tasks": failed_tasks})
    

    #[5.1] just as a backup save all arns for the given job
    #------------------------------------------------------
    for task in tasks:
        append_new_line(task_arns_file, task.id)
    
    
    
    # OLD VERSION BELOW
    
    # #  divide the circuits into 1-circuit batches (we use run_batches() and not run() because it retries failed tasks):
    # #------------------------------------------------------------------------------------------------------------
    # length = 1 #len(braket_circuits)
    # number_of_shots = metadata["number_of_repetitions"]
    # print(f"Will process {length} circuits")
    # print(f"Will process {length} circuits each one {number_of_shots} times")

    # circuits_batches = [braket_circuits[i:i + length] for i in range(0, len(braket_circuits), length)]


    # # [3] RUN
    # count = 0
    # #------------------------------------------------------------------------------------------------------------
    # # [3.1] Running batches with single circuits // tasks should be retried according to docs/Joanna
    # for batch in circuits_batches:
    #     print(f"Running  batch #{count}")
    #     count = count+1
    #      # [3.2] Run batch
    #     aws_batch = device.run_batch(task_specifications=batch,
    #                                  shots=number_of_shots,
    #                                  disable_qubit_rewiring=True)
        
    #     print(f"Created  #{len(aws_batch.tasks)} tasks for batch #{count}")


    #      # [3.3] Save task ID. In future - save results if not redundant?
    #     for task in aws_batch.tasks:
    #         print(f"task.id: {task.id}")
    #         append_new_line(task_arns_file, task.id)

    #         #Comm
    #         #results.append(task.result().measurement_counts)  #ADD
    #         #task_ids_list.append(task.id)
    #         #circuit_ids_list.append(count)

    #         pass    #endmark
    #     pass    #endmark
    
    #     results = aws_batch.result() # ALSO WILL RETRY FAILED TASKS


    # [4] Save job result - oprional, but may be heavy due to measurement_counts amount
    #save_job_result({ "measurement_counts": results, "task_id_list": task_ids_list, "circuit_id_list": circuit_ids_list})  #ADD



import copy
from typing import Optional, List, Dict, Union, Tuple

import numpy as np
from tqdm import tqdm
import seaborn as sns
import matplotlib.colors as mcolors
from matplotlib import pyplot as plt
from matplotlib import rc

from qrem.functions_qrem import functions_distances as fun_dist
from qrem.functions_qrem import povmtools
from qrem.functions_qrem.povmtools import get_enumerated_rev_map_from_indices
from qrem.noise_characterization.tomography_design.overlapping.DOTMarginalsAnalyzer import \
    DOTMarginalsAnalyzer

from qrem.common.printer import qprint

class InitialNoiseAnalyzer(DOTMarginalsAnalyzer):
    # TODO FBM, JM: think whether graphical features should be here or in some child class
    """

    """
#Mocomm - a bit wierd that som many optional POVMs
    def __init__(self,
                 results_dictionary: Dict[str, Dict[str, int]],
                 bitstrings_right_to_left: Optional[bool] = False,
                 marginals_dictionary: Optional[Dict[str, Dict[Tuple[int], np.ndarray]]] = None,
                 noise_matrices_dictionary: Optional[
                     Dict[Tuple[int], Dict[Union[str, Tuple[int]], Dict[str, np.ndarray]]]] = None,
                 clusters_list: Optional[List[Tuple[int]]] = None,
                 neighborhoods: Dict[Tuple[int], Tuple[int]] = None,
                 correlations_data: Optional[Dict[str, np.ndarray]] = None,
                 POVM_dictionary: Optional[Dict[Tuple[int], List[np.ndarray]]] = {},
                 single_qubit_errors = None,
                 qubits_indices_mapping_for_plots=None,
                 coherent_errors_1q=None,
                 correlations_pairs=None,
                 errors_data_POVMs=None,
                 coherence_data=None
                 ) -> None:

        # TODO FBM, JT: clean this up, perhaps too many arguments for initial noise analyzer

        super().__init__(results_dictionary_ddot=results_dictionary,
                         bitstrings_right_to_left=bitstrings_right_to_left,
                         marginals_dictionary=marginals_dictionary,
                         noise_matrices_dictionary=noise_matrices_dictionary
                         )

        self._correlations_data = correlations_data
        self._POVM_dictionary = POVM_dictionary
        self._errors_data_POVMs = errors_data_POVMs
        self._coherence_data = coherence_data

        if single_qubit_errors is None:
            single_qubit_errors = {(qi,): {} for qi in self._qubit_indices}

        self._single_qubit_errors = single_qubit_errors

        if clusters_list is None:
            clusters_list = []

        if neighborhoods is None:
            neighborhoods = {}

        self._clusters_list = clusters_list

        self._neighborhoods = neighborhoods

        self._coherent_errors_1q = coherent_errors_1q
        self._correlations_pairs = correlations_pairs

        fontsize_labels = 32
        fontsize_legend = 56
        fontsize_ticks = 50
        linewidth_plots = 6
        xticks_size = 20

        self._plotting_properties = {'fontsize_labels': fontsize_labels,
                                     'fontsize_legend': fontsize_legend,
                                     'fontsize_ticks': fontsize_ticks,
                                     'linewidth_plots': linewidth_plots,
                                     'xticks_size': xticks_size,
                                     'yticks_size': xticks_size
                                     # 'font_dict':{'font.size': fontsize_labels},
                                     }

        rc('text', usetex=True)
        rc('font', **{'family': 'serif',
                      'serif': ['Computer Modern'],
                      # 'weight':'bold'
                      })

        if qubits_indices_mapping_for_plots is None:
            qubits_indices_mapping_for_plots = {(qi,): qi for qi in self._qubit_indices}

        self._qubits_indices_mapping_for_plots = qubits_indices_mapping_for_plots

    @property
    def correlations_data(self):
        return self._correlations_data

    @correlations_data.setter
    def correlations_data(self, correlations_data: np.ndarray) -> None:
        self._correlations_data = correlations_data

    @property
    def errors_data(self):
        return self._errors_data_POVMs

    @errors_data.setter
    def errors_data(self, errors_data) -> None:
        self._errors_data_POVMs = errors_data


    @property
    def coherence_data(self):
        return self._coherence_data

    @coherence_data.setter
    def coherence_data(self, coherence_data) -> None:
        self._coherence_data = coherence_data







    @property
    def single_qubit_errors(self) -> dict:
        return self._single_qubit_errors

    @single_qubit_errors.setter
    def single_qubit_errors(self, single_qubit_errors: dict) -> None:
        self._single_qubit_errors = single_qubit_errors
#TODO not dectribed what this function concerns (MO) 
    def get_conservative_bound_on_statistical_errors_correlations(self,
                                                                  error_probability=0.05,
                                                                  number_of_DDOT_experiments=None):

        DDOT_keys = list(self._results_dictionary.keys())
        number_of_shots_per_setting = sum(list(self._results_dictionary[DDOT_keys[0]].values()))

        if number_of_DDOT_experiments is None:
            number_of_DDOT_experiments = len(DDOT_keys)

        number_of_qubits = self._number_of_qubits
        delta_part = np.log(1 / error_probability)
        outcomes_part = np.log(2 ** 2 - 2)
        multiple_marginals_part = np.log(number_of_qubits * (number_of_qubits - 1) * 2)

        shots_part = 2 * number_of_shots_per_setting * number_of_DDOT_experiments / 4

        return np.sqrt((delta_part + outcomes_part + multiple_marginals_part) / shots_part)

#Mocomm this method is never used - I suggest to delate it (for now I commented)


    # def _get_other_boudn_test(self,
    #                           error_probability=0.05):
    #     DDOT_keys = list(self._results_dictionary.keys())
    #     number_of_shots_per_setting = sum(list(self._results_dictionary[DDOT_keys[0]].values()))
    #     number_of_DDOT_experiments = len(DDOT_keys)

    #     shots_part = number_of_shots_per_setting * number_of_DDOT_experiments / 2

    #     outcomes_part = (1 + 1) * np.log(2)
    #     marginals_part = np.log(self._number_of_qubits)

    #     delta_part = np.log(1 / error_probability)

    #     fraction_together = -2 * (outcomes_part + marginals_part + delta_part) / shots_part
    #     # print(fraction_together, 1-fraction_together)
    #     # hej = np.
    #     epsilon = np.sqrt(np.log(1 - fraction_together))

    #     return epsilon

    # def compute_correl
#This method changes the state of "correlations data", does not return any Disctionary
    def compute_correlations_data_pairs(self,
                                       qubit_indices,
                                        distances_types=[('ac', 'classical')],
                                        chopping_threshold: Optional[float] = 0) -> Dict[
        str, Dict[str, np.array]]:

        qubit_pairs = [(qi, qj) for qi in qubit_indices for qj in qubit_indices if qj > qi]

        number_of_qubits = len(qubit_indices)

        if np.max(qubit_indices) > number_of_qubits:
            mapping = get_enumerated_rev_map_from_indices(qubit_indices)
        else:
            mapping = {qi: qi for qi in qubit_indices}

        correlations_data = {distance_tuple[0]: {} for distance_tuple in distances_types}

        for distance_tuple in distances_types:
            correlations_data[distance_tuple[0]][distance_tuple[1]] = np.zeros((number_of_qubits,
                                                                                number_of_qubits),
                                                                               dtype=float)

        for distance_tuple in distances_types:
            qprint("\nCalculating correlations of type:", distance_tuple, 'red')

            (distance_type, correlations_type) = distance_tuple

            classical_correlations = False
            if correlations_type.lower() in ['classical', 'diagonal']:
                classical_correlations = True

            for (qi, qj) in tqdm(qubit_pairs):
                qi_mapped, qj_mapped = mapping[qi], mapping[qj]

                povm_2q_now = self._POVM_dictionary[(qi, qj)]

                c_i_j, c_j_i = fun_dist.find_correlations_coefficients(povm_2q=povm_2q_now,
                                                                       distance_type=distance_type,
                                                                       classical=classical_correlations,
                                                                       direct_optimization=True)

                if c_i_j >= chopping_threshold:
                    correlations_data[distance_type][correlations_type][qi_mapped, qj_mapped] = c_i_j

                if c_j_i >= chopping_threshold:
                    correlations_data[distance_type][correlations_type][qj_mapped, qi_mapped] = c_j_i
            qprint("DONE")

        if self._correlations_data is None:
            self._correlations_data = correlations_data
        else:
            self._correlations_data = {**self._correlations_data, **correlations_data}


#JT: This function is used to compute a distance between an ideal POVM and the reconstructed one
#qubits_subsets - a list of qubits for which the computation is to be performed, for 1 qunits POVMs entries of the list consits of a tuple contaning one integer
# for two qubit POVMs tuple contains two integers
#distance types a list of tuples, contaning strings, that specifies distance(s) to be computed
#chopping_treshold - a threshold for computed distances
# resulting data structure distance_name :{distance_type:{(qubits_subset): distance_value}}


#

    def compute_errors_POVMs(self,
                             qubits_subsets: List[Tuple[int]],
                             distances_types=[('ac', 'classical')],
                             chopping_threshold: Optional[float] = 0) -> Dict[
        str, Dict[str, Dict[Tuple[int], float]]]:

        #JT: a list contaning uniqe sizes of qubits subsets passed to the class

        unique_subsets_sizes = list(np.unique([len(x) for x in qubits_subsets]))

        #JT: a dictionary contaning all ideal, projetive measurements in the computational basis
        # for dimensions corresponding to the unique dimensions passed to the input

        POVMs_compuational_basis = {x: povmtools.computational_projectors(d=2, n=x) for x in
                                    unique_subsets_sizes}

        #JT: errors_data is a dictionary that stores results of computation
        #the keys are the first entry of a tuple corresponding to chosen distances (ac/wc)
        #this is a nested dictionary - values will be also dictionaries

        errors_data = {distance_tuple[0]: {} for distance_tuple in distances_types}

        #JT: the second key in errors data is a string classical/quantum"
        #JT: We need to discuess if we want to store it this way


        for distance_tuple in distances_types:
            errors_data[distance_tuple[0]][distance_tuple[1]] = {}

        #JT: a loop to compute distances

        for distance_tuple in distances_types:



            #JT: information about computation is printed

            qprint("\nCalculating errors of type:", distance_tuple, 'red')

            #JT: an internal loop over subsets

            for subset in tqdm(qubits_subsets):

                #JT: Povm corresponding to subset is loaded into povm_now

                povm_now = self._POVM_dictionary[subset]

                #JT: function calculating specified distances, detailed comments in function distances and povmtools
                #functions involved include average_distance_POVMs and  operational_distance_POVMs

                distance_now = fun_dist.calculate_distance_between_POVMs(POVM_1=povm_now,
                                                                         POVM_2=
                                                                         POVMs_compuational_basis[
                                                                             len(subset)],
                                                                         distance_type_tuple=distance_tuple)

                #JT: treshold check

                if distance_now >= chopping_threshold:
                    errors_data[distance_tuple[0]][distance_tuple[1]][subset] = distance_now

            qprint("DONE")

        #JT: If no distances are computed the resulting dictionary is assigned to _errors_data_POVMs

        if self._errors_data_POVMs is None:
            self._errors_data_POVMs = errors_data

        #JT: a union of dictionaries is performed when some distances are computed

        else:
            self._errors_data_POVMs = {**self._errors_data_POVMs, **errors_data}

    #JT: this method is used to compute distances between coherent and their diagonal parts
    #inputs: substes_of_qubits - determines POVMs to be computed, distances_types -types of distances to be computed
    def compute_coherences_POVMs(self,
                                 qubits_subsets: List[Tuple[int]],
                                 distances_types=[('ac', 'quantum')],
                                 chopping_threshold: Optional[float] = 0) -> Dict[
        str, Dict[str, Dict[Tuple[int], float]]]:



        #JT: diagonal part of POVM is taken

        POVMs_diagonal = {subset: povmtools.get_diagonal_povm_part(self._POVM_dictionary[subset]) for
                          subset in qubits_subsets}

        #JT a nested dictionary is initialized


        coherences_data = {distance_tuple[0]: {} for distance_tuple in distances_types}

        for distance_tuple in distances_types:
            coherences_data[distance_tuple[0]][distance_tuple[1]] = {}

        #JT: start of calculation message is printed

        for distance_tuple in distances_types:
            qprint("\nCalculating errors of type:", distance_tuple, 'red')

            #JT: warning message when classical type of distance is chosen as results are trivial

            if distance_tuple[1].lower() in ['classical']:
                qprint("Warning:",
                               f"Coherence for classical part of POVM is by definition zero.", 'red')

            for subset in tqdm(qubits_subsets):
                povm_now = self._POVM_dictionary[subset]

                #JT: calculation of distances is performed in the same way as in compute_errors_POVMs

                distance_now = fun_dist.calculate_distance_between_POVMs(POVM_1=povm_now,
                                                                         POVM_2=POVMs_diagonal[subset],
                                                                         distance_type_tuple=distance_tuple)

                if distance_now >= chopping_threshold:
                    coherences_data[distance_tuple[0]][distance_tuple[1]][subset] = distance_now

            qprint("DONE")

        #JT: update of class property storing data

        if self._coherence_data is None:
            self._coherence_data = coherences_data
        else:
            self._coherence_data = {**self._coherence_data, **coherences_data}
#MOcomm - this is an old function - do we need to use it - it is used 
    def _compute_correlations_data_pairs_old(self,
                                             qubit_indices: Optional[List[int]] = None,
                                             chopping_threshold: Optional[float] = 0.) -> np.ndarray:
        """From marginal noise matrices, get correlations between pairs of qubits.
           Correlations are defined as:

           c_{j -> i_index} =
                           1/2 * || Lambda_{i_index}^{Y_j = '0'} - Lambda_{i_index}^{Y_j = '0'}||_{l1}

           Where Lambda_{i_index}^{Y_j} is an effective noise matrix on qubit "i_index"
           (averaged over all other of qubits except "j"), provided that input state
           of qubit "j" was "Y_j". Hence, c_{j -> i_index} measures how much
           noise on qubit "i_index" depends on the input state of qubit "j".

           :param qubit_indices: list of integers labeling the qubits we want to consider
                  if not provided, uses class property self._qubit_indices

           :param chopping_threshold: numerical value, for which correlations lower than
                  chopping_threshold are set to 0. If not provided, does not chop.
                  In general, it is advisable to set such cluster_threshold that
                  cuts off values below expected statistical fluctuations.

           :return: correlations_table_quantum (ARRAY):
                    element correlations_table_quantum[i_index,j] =
                    how qubit "j" AFFECTS qubit "i_index"
                    [= how noise on qubit "i_index" depends on "j"]
           """

        add_property = False
        if qubit_indices is None:
            add_property = True
            qubit_indices = self._qubit_indices

        number_of_qubits = len(qubit_indices)
        correlations_table_average = np.zeros((number_of_qubits, number_of_qubits))
        correlations_table_worst = np.zeros((number_of_qubits, number_of_qubits))

        if np.max(qubit_indices) > number_of_qubits:
            mapping = get_enumerated_rev_map_from_indices(qubit_indices)
        else:
            mapping = {qi: qi for qi in qubit_indices}

        for qi in qubit_indices:
            for qj in qubit_indices:
                qi_mapped, qj_mapped = mapping[qi], mapping[qj]
                if qj > qi:

                    # TODO FBM: ADJUST FOR FULL QDT
                    lam_i_j = self.get_noise_matrix_dependent((qi,),
                                                              (qj,))
                    lam_j_i = self.get_noise_matrix_dependent((qj,),
                                                              (qi,))

                    # how "i" is affected by "j"
                    diff_i_j = lam_i_j['0'] - lam_i_j['1']

                    # how "j" is affected by "i"
                    diff_j_i = lam_j_i['1'] - lam_j_i['0']

                    lam2q = self.get_noise_matrix_averaged(subset=(qi, qj))
                    povm2q_diagonal = povmtools.get_povm_from_stochastic_map(stochastic_map=lam2q)

                    correlation_i_j_worst = 1 / 2 * np.linalg.norm(diff_i_j, ord=1)
                    correlation_j_i_worst = 1 / 2 * np.linalg.norm(diff_j_i, ord=1)

                    # test_ij, test_ji =fun_dist.find_correlations_coefficients(povm_2q=povm2q_diagonal,
                    #                                                  distance_type='worst-case',
                    #                                                           classical=True)
                    # print()
                    # print(abs(correlation_i_j_worst-test_ij)<=10**(-8), abs(correlation_j_i_worst- test_ji)<=10**(-8))
                    #

                    correlation_i_j_average, correlation_j_i_average = fun_dist.find_correlations_coefficients(
                        povm_2q=povm2q_diagonal,
                        distance_type='average-case',
                        classical=True)

                    if correlation_i_j_worst >= chopping_threshold:
                        correlations_table_worst[qi_mapped, qj_mapped] = correlation_i_j_worst

                    if correlation_j_i_worst >= chopping_threshold:
                        correlations_table_worst[qj_mapped, qi_mapped] = correlation_j_i_worst

                    if correlation_i_j_average >= chopping_threshold:
                        correlations_table_average[qi_mapped, qj_mapped] = correlation_i_j_average

                    if correlation_j_i_average >= chopping_threshold:
                        correlations_table_average[qj_mapped, qi_mapped] = correlation_j_i_average

        if add_property:
            if 'diagonal' not in self._correlations_data.keys():
                self._correlations_data['diagonal'] = {}
            self._correlations_data['diagonal']['average-case'] = correlations_table_average
            self._correlations_data['diagonal']['worst-case'] = correlations_table_worst
        # raise KeyboardInterrupt
        return {'diagonal': {'average-case': correlations_table_average,
                             'worst-case': correlations_table_worst}}
#Mocomm we never use this method, I suggest to delate it
    def compute_1q_errors(self,
                          qubit_indices=None):
        # TODO FBM: add statistical errors
        if qubit_indices is None:
            qubit_indices = self._qubit_indices

        for qi in qubit_indices:
            noise_matrix = self.get_noise_matrix_averaged(subset=(qi,))

            p_10 = noise_matrix[1, 0]
            p_01 = noise_matrix[0, 1]

            self._single_qubit_errors[(qi,)]['1|0'] = p_10
            self._single_qubit_errors[(qi,)]['0|1'] = p_01

        return self._single_qubit_errors


#MOcomm - not sure if needed I suggest delating
#    def print_properties(self):
#        # TODO FBM, OS: add this
#
#        return None

#MOcomm - not sure if needed I suggest delating
#    def draw_noise_model(self):
#        # TODO FBM, JM: add this
#
#        return None



#####  BELOW WE HAWE SOME CRAZY DROWING TOOLS - THEY SHOULD NOT BE HERE!!!! (MO)
# ORGANIZE - everything below is research-specific and we don't use it anymore. A basic histogram is implemented in
# visualisation.ResultsForPlotting method draw_histogram, the research scripts are in use_cases. This can be deleted.

    def __prepare_drawing_1q_errors_barplot(self,
                                            axis,
                                            errors_list,
                                            which_error: str):

        if which_error == "p(0|1)":
            error_insert_string = r"p(0$\vert$1)"
            color = 'red'
        elif which_error == "p(1|0)":
            error_insert_string = r"p(1$\vert$0)"
            color = 'green'

        ylabel = fr"Error prob. {error_insert_string} [$\%$]"

        qubits_labels = [f"{self._qubits_indices_mapping_for_plots[(qi,)]}" for qi in
                         self._qubit_indices]
        bar_width = self._number_of_qubits / 10 * 0.6
        labelpad_x = bar_width * 1.5

        axis.set_xlabel('Qubit label', fontsize=self._plotting_properties['fontsize_labels'])
        axis.set_ylabel(ylabel,
                        fontsize=self._plotting_properties['fontsize_labels'])
        axis.bar(x=self._qubit_indices,
                 height=errors_list,
                 width=bar_width,
                 color=color)
        axis.xaxis.labelpad = labelpad_x
        axis.set_xticks(self._qubit_indices)
        axis.set_xticklabels(qubits_labels,
                             rotation=30,
                             fontsize=self._plotting_properties['xticks_size'])

        mean_errors = np.mean(errors_list)
        position_right_mean = self._number_of_qubits - 0.5

        axis.set_xlim(-1, self._number_of_qubits)

        max_ytick = max([np.round(x, 1) for x in errors_list])
        axis.set_ylim(0, max_ytick * 1.05)

        axis.hlines(y=mean_errors,
                    xmin=-0.5,
                    xmax=position_right_mean,
                    linestyles='--',
                    colors='black',
                    label='mean',
                    linewidth=3)

        axis.annotate(f"mean = {np.round(mean_errors, 1)}",
                      xy=(position_right_mean - 1.25, mean_errors + 0.05 * max_ytick),
                      fontsize=self._plotting_properties['fontsize_labels'])

        #

        for tick in axis.yaxis.get_major_ticks():
            tick.label.set_fontsize(self._plotting_properties['yticks_size'])

        return axis
    #ORGANIZE - delete
    def __draw_1q_errors_barplot(self,
                                 additional_string_title=None):

        # qubits_mapping = self._qubits_indices_mapping_for_plots

        errors_10 = [self._single_qubit_errors[(qi,)]['1|0'] * 100
                     for qi in self._qubit_indices]
        errors_01 = [self._single_qubit_errors[(qi,)]['0|1'] * 100
                     for qi in self._qubit_indices]
        figure_pyplot = plt.figure()
        axis_10 = figure_pyplot.add_subplot(2, 1, 1)
        axis_10 = self.__prepare_drawing_1q_errors_barplot(axis=axis_10,
                                                           which_error="p(1|0)",
                                                           errors_list=errors_10)

        axis_01 = figure_pyplot.add_subplot(2, 1, 2)
        axis_01 = self.__prepare_drawing_1q_errors_barplot(axis=axis_01,
                                                           which_error="p(0|1)",
                                                           errors_list=errors_01)

        figure_title = "1q errors"

        if additional_string_title is not None:
            figure_title = figure_title + f"{additional_string_title}"
        figure_pyplot.suptitle(figure_title, fontsize=self._plotting_properties['fontsize_labels'],
                               y=0.98)
        figure_pyplot.tight_layout()

        return figure_pyplot

#ORGANIZE MO - for sure plotting does not belong here (MO)! ###########
#ORGANIZE JM - I agree, let's delete this
    def __prepare_drawing_1q_errors_histogram(self,
                                              axis,
                                              errors_list,
                                              which_error: str,
                                              cumulative=False):

        if which_error == "p(0|1)":
            error_insert_string = r"p(0$\vert$1)"
            color = 'red'
        elif which_error == "p(1|0)":
            error_insert_string = r"p(1$\vert$0)"
            color = 'green'
        elif which_error.upper() == 'AVERAGE-CASE DISTANCE':
            color = 'green'
            error_insert_string = r"coherent average-case distance"

        elif which_error.upper() == 'WORST-CASE DISTANCE':
            color = 'red'
            error_insert_string = r"coherent worst-case distance"

        # ylabel = fr"Error prob. {error_insert_string} [$\%$]"
        #
        #
        # qubits_labels = [f"{self._qubits_indices_mapping_for_plots[(qi,)]}" for qi in
        #                  self._qubit_indices]

        #
        # bar_width = self._number_of_qubits / 10 * 0.6
        # labelpad_x = bar_width * 1.5
        # error_insert_string = 'coherent'
        axis.set_xlabel(f'Error {error_insert_string} [\%]',
                        fontsize=self._plotting_properties['fontsize_labels'])
        # TODO FBM: putting by hand
        # axis.set_xlabel(f'Coherent error [\%]', fontsize=self._plotting_properties['fontsize_labels'])

        if cumulative:
            axis.hist(
                x=errors_list,
                bins=int(self._number_of_qubits),
                color=color,
                cumulative=cumulative,
                density=True,
                histtype='step',
                # stacked=True
                # x=self._qubit_indices,
                #      height=errors_list,
                #      width=bar_width,
                #      color=color
            )

            axis.set_ylabel("CDF",
                            fontsize=self._plotting_properties['fontsize_labels'])

            max_error = np.max(errors_list)

            axis.vlines(x=max_error,
                        ymin=0.,
                        ymax=1,
                        linestyles='-',
                        colors='white',
                        # label='mean',
                        linewidth=3)






        else:

            axis.set_ylabel("Amount of qubits",
                            fontsize=self._plotting_properties['fontsize_labels'])

            axis.hist(
                x=errors_list,
                bins=int(self._number_of_qubits),
                color=color,
                cumulative=cumulative
                # x=self._qubit_indices,
                #      height=errors_list,
                #      width=bar_width,
                #      color=color
            )
        # axis.xaxis.labelpad = labelpad_x

        min, max, mean = int(np.floor(np.min(errors_list))), int(
            np.floor(np.max(errors_list))), np.mean(errors_list)

        axis.set_xticks(list(np.round(np.linspace(min, mean, 4), 1)) +
                        list(np.round(np.linspace(mean, max * 1.4, 8), 1)))

        # axis.set_xticks([0.5, 1, 2, 4, 5] + list(range(10,max_error+1,5)))
        # axis.set_xticklabels(qubits_labels,
        #                      rotation=30,
        #                      fontsize=self._plotting_properties['xticks_size'])

        # mean_errors = np.mean(errors_list)
        # position_right_mean = mean_errors

        # axis.set_xlim(-1, self._number_of_qubits)

        # max_xtick = max([np.round(x, 1) for x in errors_list])
        # axis.set_xlim(0, max_xtick * 1.05)

        # axis.hlines(y = mean_errors,
        #             xmin= -0.5,
        #             xmax = max_xtick,
        #             linestyles='--',
        #             colors='black',
        #             label='mean',
        #             linewidth=3)
        #
        # axis.annotate(f"mean = {np.round(mean_errors,1)}",
        #               xy=(position_right_mean-1.25, mean_errors+0.05*max_xtick),
        #               fontsize=self._plotting_properties['fontsize_labels'])
        #
        #
        # #

        for tick in axis.yaxis.get_major_ticks():
            tick.label.set_fontsize(self._plotting_properties['yticks_size'])

        return axis
    #ORGANIZE - delete
    def __draw_errors_histogram(self,
                                additional_string_title=None,
                                errors_type='classical',
                                histogram_type=None):

        if histogram_type is None:
            if self._number_of_qubits > 30:
                histogram_type = 'cumulative'
            else:
                histogram_type = 'bar'

        if histogram_type.upper() == 'BAR':
            cumulative_histogram = False
        elif histogram_type.upper() == 'CUMULATIVE':
            cumulative_histogram = True

        # qubits_mapping = self._qubits_indices_mapping_for_plots

        if errors_type.upper() == 'CLASSICAL':
            errors_type_0_list = [self._single_qubit_errors[(qi,)]['1|0'] * 100
                                  for qi in self._qubit_indices]
            errors_type_1_list = [self._single_qubit_errors[(qi,)]['0|1'] * 100
                                  for qi in self._qubit_indices]

            errors_type_0_name = "p(1|0)"
            errors_type_1_name = "p(0|1)"

        elif errors_type.upper() == 'COHERENT':
            errors_type_0_list = [self._coherent_errors_1q[(qi,)]['average_case'] * 100
                                  for qi in self._qubit_indices]
            errors_type_1_list = [self._coherent_errors_1q[(qi,)]['worst_case'] * 100
                                  for qi in self._qubit_indices]

            errors_type_0_name = 'average-case distance'
            errors_type_1_name = 'worst-case distance'

        # elif err

        figure_pyplot = plt.figure()
        axis_10 = figure_pyplot.add_subplot(1, 1, 1)
        axis_10 = self.__prepare_drawing_1q_errors_histogram(axis=axis_10,
                                                             which_error=errors_type_0_name,
                                                             errors_list=errors_type_0_list,
                                                             cumulative=cumulative_histogram)

        # axis_01 = figure_pyplot.add_subplot(2, 1, 2)
        # axis_01 = self.__prepare_drawing_1q_errors_histogram(axis=axis_01,
        #                                                    which_error=errors_type_1_name,
        #                                                    errors_list=errors_type_1_list,
        #                                                      cumulative=cumulative_histogram)

        figure_title = "1q errors"

        if additional_string_title is not None:
            figure_title = figure_title + f"{additional_string_title}"
        # figure_pyplot.suptitle(figure_title, fontsize=self._plotting_properties['fontsize_labels'], y=0.98)
        figure_pyplot.tight_layout()

        return figure_pyplot
    #ORGANIZE - delete
    def draw_1q_errors(self,
                       additional_string_title=None,
                       plottype=None,
                       errors_type='classical'):

        additional_string_title = additional_string_title.replace('_', '-')

        if plottype is None:
            if self._number_of_qubits <= 25:
                figure_pyplot = self.__draw_1q_errors_barplot(
                    additional_string_title=additional_string_title,
                    errors_type=errors_type)

            else:
                figure_pyplot = self.__draw_errors_histogram(additional_string_title=
                                                             additional_string_title,
                                                             errors_type=errors_type)
        else:
            # TODO FBM: make it smarter
            if plottype.upper() == 'BARPLOT':
                figure_pyplot = self.__draw_1q_errors_barplot(
                    additional_string_title=additional_string_title,
                    errors_type=errors_type)
            elif plottype.upper() == "HISTOGRAM-BAR":
                figure_pyplot = self.__draw_errors_histogram(
                    additional_string_title=additional_string_title,
                    errors_type=errors_type,
                    histogram_type='BAR')
            elif plottype.upper() == "HISTOGRAM-CUMULATIVE":
                figure_pyplot = self.__draw_errors_histogram(
                    additional_string_title=additional_string_title,
                    errors_type=errors_type,
                    histogram_type='CUMULATIVE')

        return figure_pyplot

    #ORGANIZE: this can be deleted, it's only used in draw_correlations below, which can be deleted
    def __prepare_correlations_drawing_heatmap_classical(self,
                                                         additional_string_title=None,
                                                         ):

        title_heatmap = f"Correlations table"

        if additional_string_title is not None:
            title_heatmap = title_heatmap + f"{additional_string_title}"

        correlations_table = copy.deepcopy(self._correlations_table)

        for i in range(self._number_of_qubits):
            for j in range(self._number_of_qubits):
                if correlations_table[i, j] != 0:
                    correlations_table[i, j] *= 100

        statistics_bound = self.get_conservative_bound_on_statistical_errors_correlations() * 100
        max_value = correlations_table.max()

        percentage_cutoff = statistics_bound / max_value
        percentage_notcutoff = 1 - percentage_cutoff

        number_of_points_cutoff = int(np.floor(percentage_cutoff * 128))
        number_of_points_notcutoff = int(np.ceil(percentage_notcutoff * 128))
        if number_of_points_cutoff == 0:
            number_of_points_cutoff += 1
            number_of_points_notcutoff += -1

        color_zero = plt.cm.nipy_spectral_r([0])
        colors_list_zero = np.array([list(color_zero[0]) for _ in range(number_of_points_cutoff)])
        colors_correlations = plt.cm.gist_heat_r(np.linspace(0, 1, number_of_points_notcutoff))[
                              int(number_of_points_notcutoff * 0.1):]

        colors_merged = np.vstack((colors_list_zero, colors_correlations))

        color_map_merged = mcolors.LinearSegmentedColormap.from_list('colormap_merged', colors_merged)

        heatmap = sns.heatmap(correlations_table,
                              cmap=color_map_merged,
                              # norm=divnorm,
                              linewidths=1,
                              linecolor='white',
                              cbar=True,
                              # cbar_kws={'label': r"$c_{i \rightarrow j}\ [\%]$"}
                              )

        ticks_fontsize = self._plotting_properties['fontsize_labels']
        if self._number_of_qubits > 30:
            ticks_fontsize *= 1 / 3

        # if self._number_of_qubits > 80:
        #     ticks_fontsize*=1/2

        y_label = fr"Affected qubit label"
        x_label = fr"Affecting qubit label"

        qubits_labels = [f"{self._qubits_indices_mapping_for_plots[(qi,)]}" for qi in
                         self._qubit_indices]

        ###########################################################################################
        ticks = [i + 0.5 for i in range(len(qubits_labels))]
        heatmap.set_yticks(ticks)
        heatmap.set_xticks(ticks)

        heatmap.set_xticklabels(qubits_labels,
                                rotation=90,
                                fontsize=ticks_fontsize)

        heatmap.set_yticklabels(qubits_labels,
                                fontsize=ticks_fontsize,
                                rotation=0)

        ###########################################################################################

        bar_width = self._number_of_qubits / 10 * 0.6
        # labelpad_x = bar_width * 1.5

        heatmap.set_xlabel(x_label, fontsize=ticks_fontsize)
        heatmap.set_ylabel(y_label,
                           fontsize=ticks_fontsize)

        cbar = heatmap.collections[0].colorbar
        cbar.set_label(r"$c_{i \rightarrow j}\ [\%]$",
                       labelpad=50,
                       rotation=0,
                       fontsize=self._plotting_properties['fontsize_ticks'])
        # cbar.set_annotation("HEJKA")

        cbar.ax.tick_params(labelsize=self._plotting_properties['fontsize_labels'])
        yticks = cbar.ax.get_yticks()
        yticks_labels = list(cbar.ax.get_yticklabels())

        cbar.ax.annotate(r"$\leftarrow$ cutoff",
                         (1., percentage_cutoff * 1.02),
                         fontsize=self._plotting_properties['fontsize_labels'])

        #
        #
        # print(yticks)
        # print(yticks_labels)
        # raise KeyboardInterrupt

        # cbar.ax.set_yticklabels()

        heatmap.set_title(title_heatmap,
                          fontsize=self._plotting_properties['fontsize_labels'],
                          y=0.98)
        # heatmap.xaxis.tick_top()
        # heatmap.xaxis.set_label_position('top')

        return heatmap
    #ORGANIZE: this can be deleted, it's only used in draw_correlations below, which can be deleted
    def __prepare_correlations_drawing_heatmap_nonclassical(self,
                                                            additional_string_title=None,
                                                            correlations_type='COHERENT-AVERAGE'
                                                            ):

        title_heatmap = f"Correlations table, {correlations_type}"

        if additional_string_title is not None:
            title_heatmap = title_heatmap + f"\n{additional_string_title}"

        correlations_table = np.zeros((self._number_of_qubits, self._number_of_qubits))

        for i in range(self._number_of_qubits):
            for j in range(i + 1, self._number_of_qubits):

                if correlations_type.upper() == 'COHERENT-AVERAGE':
                    corr_now = self._correlations_pairs['coherent'][(i, j)]['average']
                elif correlations_type.upper() == 'COHERENT-WORST':
                    corr_now = self._correlations_pairs['coherent'][(i, j)]['worst']

                if corr_now != 0:
                    correlations_table[i, j] = 100 * corr_now

        max_value = correlations_table.max()
        percentage_cutoff = 0
        percentage_notcutoff = 1 - percentage_cutoff
        number_of_points_cutoff = int(np.floor(percentage_cutoff * 128))
        number_of_points_notcutoff = int(np.ceil(percentage_notcutoff * 128))
        if number_of_points_cutoff == 0:
            number_of_points_cutoff += 1
            number_of_points_notcutoff += -1

        color_zero = plt.cm.nipy_spectral_r([0])
        colors_list_zero = np.array([list(color_zero[0]) for _ in range(number_of_points_cutoff)])
        colors_correlations = plt.cm.gist_heat_r(np.linspace(0, 1, number_of_points_notcutoff))[
                              int(number_of_points_notcutoff * 0.1):]

        colors_merged = np.vstack((colors_list_zero, colors_correlations))

        color_map_merged = mcolors.LinearSegmentedColormap.from_list('colormap_merged', colors_merged)

        heatmap = sns.heatmap(correlations_table,
                              cmap=color_map_merged,
                              # norm=divnorm,
                              linewidths=5,
                              linecolor='white',
                              cbar=True,
                              # cbar_kws={'label': r"$c_{i \rightarrow j}\ [\%]$"}
                              )

        y_label = fr"qubit label"
        x_label = fr"qubit label"

        qubits_labels = [f"{self._qubits_indices_mapping_for_plots[(qi,)]}" for qi in
                         self._qubit_indices]
        ###########################################################################################
        ticks = [i + 0.5 for i in range(len(qubits_labels))]
        heatmap.set_yticks(ticks)
        heatmap.set_xticks(ticks)

        ticks_fontsize = self._plotting_properties['fontsize_labels']

        if self._number_of_qubits > 30:
            ticks_fontsize *= 1 / 3

        heatmap.set_xticklabels(qubits_labels,
                                rotation=30,
                                fontsize=ticks_fontsize)

        heatmap.set_yticklabels(qubits_labels,
                                fontsize=ticks_fontsize,
                                rotation=0)
        ###########################################################################################

        bar_width = self._number_of_qubits / 10 * 0.6
        # labelpad_x = bar_width * 1.5

        heatmap.set_xlabel(x_label, fontsize=self._plotting_properties['fontsize_labels'])

        heatmap.set_ylabel(y_label,
                           fontsize=self._plotting_properties['fontsize_labels'])

        cbar = heatmap.collections[0].colorbar
        cbar.set_label(r"$[\%]$",
                       labelpad=50,
                       rotation=0,
                       fontsize=self._plotting_properties['fontsize_ticks'])

        cbar.ax.tick_params(labelsize=self._plotting_properties['fontsize_labels'])

        heatmap.set_title(title_heatmap,
                          fontsize=self._plotting_properties['fontsize_labels'],
                          y=0.98)

        return heatmap
    #ORGANIZE: same functionality is implemented in visualisation.ResultsForPlotting method draw_heatmap
    #This can be deleted
    def draw_correlations(self,
                          additional_string_title=None,
                          correlations_type='classical',
                          plottype='heatmap'):

        additional_string_title = additional_string_title.replace('_', '-')

        # TODO FBM, JM: Think whether to do some more plot options

        if plottype.lower() in ['heatmap']:

            if correlations_type.upper() == 'CLASSICAL':
                plot = self.__prepare_correlations_drawing_heatmap_classical(
                    additional_string_title=additional_string_title)
            else:
                plot = self.__prepare_correlations_drawing_heatmap_nonclassical(
                    additional_string_title=additional_string_title,
                    correlations_type=correlations_type)


        elif plottype.lower() in ['histogram']:
            if correlations_type.upper() == 'CLASSICAL':
                raise KeyError()
                histogram = self.draw_1q_errors(
                    additional_string_title=None,
                    plottype=None,
                    errors_type='classical')

            else:
                raise ValueError()
                plot = self.__prepare_correlations_drawing_heatmap_nonclassical(
                    additional_string_title=additional_string_title,
                    correlations_type=correlations_type)

        return plot




from typing import Tuple, Dict, List, Union, Optional

import numpy as np

from qrem.functions_qrem import ancillary_functions as anf, functions_data_analysis as fda
from qrem.functions_qrem import quantum_ancillary_functions as qanf 
from qrem.noise_characterization.base_classes.marginals_analyzer_base import MarginalsAnalyzerBase
from qrem.noise_model_generation.CN.GlobalNoiseMatrixCreator import GlobalNoiseMatrixCreator

from qrem.common import convert, utils

#TODO FBM: Clean up this file


# {cluster_tuple:{'averaged'}}









##############################################################
##################### Noise model sampling##########
##############################################################

def probabilistic_bitflip_choices(bit: str,
                                  bitflip_probability: float):
    # this is faster than using np.random.binomial
    import random
    return random.choices([convert.negate_bitstring(bit), bit],
                          weights=[bitflip_probability, 1 - bitflip_probability])[0]


def stochastic_relabeling(bitstring: str,
                          stochastic_matrices: Dict[str, np.ndarray],
                          reverse_bitstring=False):
    number_of_qubits = len(list(bitstring))

    flipped_bitstring = ''
    #
    if reverse_bitstring:
        bitstring = bitstring[::-1]

    for qubit_index in range(number_of_qubits):
        bit = bitstring[qubit_index]
        stochastic_map = stochastic_matrices['q%s' % qubit_index]
        bitflip_probability = stochastic_map[int(convert.negate_bitstring(bit), 2), int(bit, 2)]
        flipped_bit = probabilistic_bitflip_choices(bit, bitflip_probability)
        flipped_bitstring += flipped_bit

    return flipped_bitstring



def sample_from_noise_model(input_state: List[str],
                            number_of_qubits: int,
                            number_of_samples: int,
                            noise_model_description: Dict[Tuple[int], Dict[str, np.ndarray]],
                            neighborhoods_tuples: Dict[Tuple[int], List[int]]
                            ):
    noisy_bitstrings = [np.zeros((number_of_qubits), dtype=str) for _ in range(number_of_samples)]
    for cluster_qubits, noise_matrices_on_cluster in noise_model_description.items():
        # print(neighborhoods_tuples[cluster_qubits])
        # print(cluster_qubits)
        # print(cluster_qubits)
        input_on_cluster = ''.join([input_state[x] for x in cluster_qubits])

        neighbors = neighborhoods_tuples[cluster_qubits]

        if len(neighbors) == 0:
            noise_matrix = noise_matrices_on_cluster['averaged']
        else:
            input_on_neighbors = ''.join(
                [input_state[x] for x in neighbors])

            noise_matrix = noise_matrices_on_cluster[neighbors][input_on_neighbors]

        prob_distro_now = noise_matrix[:, int(input_on_cluster, 2)]
        cluster_qubits_enumerated = dict(enumerate(cluster_qubits))
        # print(noise_matrix)
        for sample_index in range(number_of_samples):
            resampling = np.random.multinomial(n=1, pvals=prob_distro_now)
            new_bitstring_marginal = list(convert.integer_to_bitstring(integer=np.argmax(resampling),
                                                                   number_of_bits=len(cluster_qubits)))

            for ordered_index, true_index in cluster_qubits_enumerated.items():
                noisy_bitstrings[sample_index][true_index] = new_bitstring_marginal[ordered_index]

    return [list(bitstring) for bitstring in noisy_bitstrings]





















##############################################################
##################### Model simulation##########
##############################################################


def construct_product_noise_model_in_correct_order(
        noise_matrices_dictionary: Dict[Tuple[int], np.ndarray],
        qubits_mapping_dictionary: Optional[Dict[int, int]] = None):
    all_clusters = list(noise_matrices_dictionary.keys())

    # print(all_clusters)
    # print([list(x) for x in all_clusters])
    number_of_qubits = len(utils.lists_sum_multiple([list(x) for x in all_clusters]))

    if qubits_mapping_dictionary is None:
        def qubits_map(x: int):
            return x
    else:
        def qubits_map(x: int):
            return qubits_mapping_dictionary[x]

    global_noise_matrix = np.eye(2 ** number_of_qubits, dtype=float)

    for cluster, noise_matrix in noise_matrices_dictionary.items():
        embeded_noise_matrix = qanf.embed_operator_in_bigger_hilbert_space(
            number_of_qubits=number_of_qubits,
            local_operator=noise_matrix,
            global_indices=[qubits_map(qi) for qi in cluster]).real

        global_noise_matrix = global_noise_matrix @ embeded_noise_matrix

    return global_noise_matrix



def get_random_clusters_division(number_of_qubits: int,
                                 minimal_cluster_size: int,
                                 maximal_cluster_size: int):
    clusters_list = []
    assigned_qubits = []
    while sorted(assigned_qubits) != list(range(number_of_qubits)):

        number_of_qubits_left = number_of_qubits - len(assigned_qubits)

        if number_of_qubits_left < maximal_cluster_size:
            maximal_cluster_size = number_of_qubits_left

        cluster_size = np.random.randint(minimal_cluster_size, maximal_cluster_size + 1)
        cluster_now = []

        while len(cluster_now) != cluster_size:
            random_qubit = np.random.randint(0, number_of_qubits)

            if random_qubit not in assigned_qubits:
                cluster_now.append(random_qubit)
                assigned_qubits.append(random_qubit)

        clusters_list.append(tuple(sorted(cluster_now)))

    return clusters_list





def get_noisy_energy_exact_CN(input_state,
                              noise_matrices_dictionary_tuples,
                              neighborhoods_dictionary_tuples,
                              weights_dictionary_tuples
                              ):
    noise_model_now = get_model_for_fixed_input_state(input_state=input_state,
                                                      noise_model_description=noise_matrices_dictionary_tuples,
                                                      neighborhoods_dictionary=neighborhoods_dictionary_tuples)

    pairs_marginals_noisy = calculate_pairs_marginals_from_tensor_model_fixed_input(
        input_state=input_state,
        noise_matrices_dictionary=noise_model_now,
        get_also_1q_marginals=True)

    energy_modeled_exact = fda.estimate_energy_from_marginals(
        weights_dictionary=weights_dictionary_tuples,
        marginals_dictionary=pairs_marginals_noisy)

    return energy_modeled_exact


##############################################################
##################### Noise model reconstruction ##########
##############################################################

def get_noisy_energy_product_noise_model(input_state,
                                         noise_matrices_dictionary,
                                         weights_dictionary_tuples,
                                         needed_pairs=None):
    pairs_marginals_noisy = calculate_pairs_marginals_from_tensor_model_fixed_input(
        input_state=input_state,
        noise_matrices_dictionary=noise_matrices_dictionary,
        needed_pairs=needed_pairs,
        get_also_1q_marginals=True)

    energy_modeled_exact = fda.estimate_energy_from_marginals(
        weights_dictionary=weights_dictionary_tuples,
        marginals_dictionary=pairs_marginals_noisy)

    return energy_modeled_exact



#JT: This is a function used when neighbors are used (it specifies noise matrix) 

def get_model_for_fixed_input_state(input_state: List[str],
                                    noise_model_description: Dict[Tuple[int],
                                                                  Dict[Union[Tuple[int], str],
                                                                       Union[np.ndarray, Dict[
                                                                           str, np.ndarray]]
                                                                  ]
                                    ],
                                    neighborhoods_dictionary: Dict[Tuple[int],
                                                                   Tuple[int]]):
    noise_matrices_now = {}
    for cluster_qubits, noise_matrices_on_cluster in noise_model_description.items():
        neighbors_now = neighborhoods_dictionary[cluster_qubits]

        # print(cluster_qubits,neighbors_now)
        #
        # print(noise_matrices_on_cluster)

        if len(neighbors_now) == 0:
            noise_matrix = noise_matrices_on_cluster['averaged']
        else:
            input_on_neighbors = ''.join([input_state[qi] for qi in neighbors_now])
            noise_matrix = noise_matrices_on_cluster[neighbors_now][input_on_neighbors]

        noise_matrices_now[cluster_qubits] = noise_matrix

    return noise_matrices_now


##############################################################
##################### Functions computing marginals ##########
##############################################################


def get_noisy_marginals_within_cluster(input_state: str,
                                       cluster_noise_matrix: np.ndarray,
                                       subsets_of_interest: List[Union[Tuple[int], List[int]]],
                                       qubits_mapping: Dict[int, int]) -> Dict[Tuple[int], np.ndarray]:
    input_state_binary = int(input_state, 2)

    probability_distribution = cluster_noise_matrix[:, input_state_binary]

    marginals_now = {}
    for qubits_of_interest in subsets_of_interest:
        marginal = fda.get_marginal_from_probability_distribution(
            global_probability_distribution=probability_distribution,
            bits_of_interest=[qubits_mapping[x] for x in qubits_of_interest])
        marginals_now[tuple(qubits_of_interest)] = marginal

    return marginals_now


def calculate_1q_marginals_from_tensor_model(input_state: str,
                                             noise_matrices_dictionary: Dict[Tuple[int],
                                                                             np.ndarray]):
    marginals_1q = {}
    for cluster, noise_matrix in noise_matrices_dictionary.items():
        cluster_input_state = ''.join([input_state[x] for x in cluster])
        enumerated_qubits = utils.map_index_to_order(list(cluster))
        marginals_now = get_noisy_marginals_within_cluster(input_state=cluster_input_state,
                                                           cluster_noise_matrix=noise_matrix,
                                                           subsets_of_interest=[[qi] for qi in
                                                                                cluster],
                                                           qubits_mapping=enumerated_qubits)

        marginals_1q = {**marginals_1q, **marginals_now}

    return marginals_1q



def calculate_pairs_marginals_from_tensor_model_all_inputs(
        noise_matrices_dictionary: Dict[Tuple[int],
                                        np.ndarray],
        needed_pairs,
        get_also_1q_marginals=True
):
    # TODO FBM: generalize to higher locality

    clusters_list = list(noise_matrices_dictionary.keys())
    number_of_qubits = sum([len(x) for x in clusters_list])

    qubits_to_clusters_map = {}
    for qubit in range(number_of_qubits):
        for cluster in clusters_list:
            if qubit in cluster:
                qubits_to_clusters_map[qubit] = cluster

    pairs_marginals = {pair: {} for pair in needed_pairs}
    if get_also_1q_marginals:
        qubits_list = utils.lists_sum_multiple(clusters_list)
        pairs_marginals = {**pairs_marginals,
                           **{(q0,): {} for q0 in qubits_list}}

    already_done = []
    for cluster, noise_matrix in noise_matrices_dictionary.items():
        pairs_in_cluster = [(qi, qj) for qi in cluster for qj in cluster
                            if qj > qi and (qi, qj) in needed_pairs]

        local_classical_register = anf.get_classical_register_bitstrings(
            qubit_indices=range(len(cluster)))
        enumerated_qubits = utils.map_index_to_order(list(cluster))
        for cluster_input_state in local_classical_register:

            marginals_dict_now = get_noisy_marginals_within_cluster(input_state=cluster_input_state,
                                                                    cluster_noise_matrix=noise_matrix,
                                                                    subsets_of_interest=pairs_in_cluster,
                                                                    qubits_mapping=enumerated_qubits)

            for pair_now, marginal_now in marginals_dict_now.items():
                pairs_marginals[pair_now][cluster_input_state] = marginal_now

            if get_also_1q_marginals:
                marginals_dict_now1q = get_noisy_marginals_within_cluster(
                    input_state=cluster_input_state,
                    cluster_noise_matrix=noise_matrix,
                    subsets_of_interest=[[qi] for qi in cluster],
                    qubits_mapping=enumerated_qubits)
                for qubit_now, marginal1q in marginals_dict_now1q.items():
                    pairs_marginals[qubit_now][cluster_input_state] = marginal1q

        # raise KeyboardInterrupt
        already_done = already_done + pairs_in_cluster

    for (q0, q1) in utils.lists_difference(needed_pairs, already_done):

        cl0, cl1 = qubits_to_clusters_map[q0], qubits_to_clusters_map[q1]

        enumerated_qubits = utils.map_index_to_order(list(cl0) + list(cl1))

        local_classical_register = anf.get_classical_register_bitstrings(
            qubit_indices=range(len(cl0 + cl1)))

        for input_bitstring in local_classical_register:
            in_cl0 = ''.join([input_bitstring[enumerated_qubits[x]] for x in cl0])
            in_cl1 = ''.join([input_bitstring[enumerated_qubits[x]] for x in cl1])
            marginal_0 = pairs_marginals[(q0,)][in_cl0]
            marginal_1 = pairs_marginals[(q1,)][in_cl1]
            pairs_marginals[(q0, q1)][input_bitstring] = np.kron(marginal_0,
                                                                 marginal_1)

    return pairs_marginals


def calculate_pairs_marginals_from_tensor_model_fixed_input(input_state: str,
                                                            noise_matrices_dictionary: Dict[Tuple[int],
                                                                                            np.ndarray],
                                                            needed_pairs,
                                                            get_also_1q_marginals=True
                                                            ):
    # TODO FBM: generalize to higher locality
    pairs_marginals = {}
    for cluster, noise_matrix in noise_matrices_dictionary.items():
        pairs_in_cluster = [(qi, qj) for qi in cluster for qj in cluster
                            if qj > qi and (qi, qj) in needed_pairs]
        cluster_input_state = ''.join([input_state[x] for x in cluster])
        enumerated_qubits = utils.map_index_to_order(list(cluster))
        marginals_now = get_noisy_marginals_within_cluster(input_state=cluster_input_state,
                                                           cluster_noise_matrix=noise_matrix,
                                                           subsets_of_interest=pairs_in_cluster,
                                                           qubits_mapping=enumerated_qubits)

        pairs_marginals = {**pairs_marginals, **marginals_now}

        if get_also_1q_marginals:
            marginals_now = get_noisy_marginals_within_cluster(input_state=cluster_input_state,
                                                               cluster_noise_matrix=noise_matrix,
                                                               subsets_of_interest=[[qi] for qi in
                                                                                    cluster],
                                                               qubits_mapping=enumerated_qubits)

            pairs_marginals = {**pairs_marginals, **marginals_now}

    already_done = list(pairs_marginals.keys())
    for pair in utils.lists_difference(needed_pairs, already_done):
        # if pair not in already_done:
        (q0, q1) = pair
        marginal_0 = pairs_marginals[(q0,)]
        marginal_1 = pairs_marginals[(q1,)]

        pairs_marginals[(q0, q1)] = np.kron(marginal_0, marginal_1)

    return pairs_marginals



# def calculate_marginals_from_noise_model(input_state: List[str],
#                                          noise_model_description,
#                                          neighborhoods_tuples,
#                                          # number_of_qubits: int,
#                                          needed_subsets: List[str],
#                                          correction_indices: Dict[str, str],
#                                          # noise_model_description: Dict[
#                                          #     Tuple[int], Dict[str, np.ndarray]],
#                                          # neighborhoods_tuples: Dict[Tuple[int], List[int]],
#                                          clusters_dictionary: Dict[int, List[int]],
#                                          neighborhoods: Dict[Tuple[int], List[int]],
#                                          noise_matrices_dictionary: dict
#                                          ):
#     noise_matrices_now = {}
#     clusters_list = []
#     for cluster_qubits, noise_matrices_on_cluster in noise_model_description.items():
#         clusters_list.append(cluster_qubits)
#         # print(neighborhoods_tuples[cluster_qubits])
#         input_on_cluster = ''.join([input_state[x] for x in cluster_qubits])

#         if len(neighborhoods_tuples[cluster_qubits]) == 0:
#             noise_matrix = noise_matrices_on_cluster['averaged']
#         else:
#             input_on_neighbors = ''.join(
#                 [input_state[x] for x in neighborhoods_tuples[cluster_qubits]])
#             noise_matrix = noise_matrices_on_cluster[input_on_neighbors]

#         noise_matrices_now[cluster_qubits] = noise_matrix
#         # prob_distro_now = noise_matrix[:, int(input_on_cluster, 2)]

# #Mocomm - why private function inside a function...
#     def _find_clusters_in_subset(qubits_list): 
#         clusters_here = []
#         for qubit in qubits_list:
#             for cl in clusters_list:
#                 if qubit in cl:
#                     if cl not in clusters_here:
#                         clusters_here.append(cl)

#         return clusters_here
#     #TODO: undestand the role of MarginalAnalyzer below
#     marginals_analyzer = MarginalsAnalyzerBase({},
#                                                False)
                                    

#     marginal_probs_big = {}
#     marginals = {}
#     for subset in needed_subsets:
#         qubits_string = 'q' + 'q'.join([str(x) for x in subset])
#         bigger_subset = convert.get_qubit_indices_from_keystring(correction_indices[qubits_string])

#         clusters_now = _find_clusters_in_subset(bigger_subset)

#         # qubits_ordering_clusters = []
#         qubits_ordering_all = []
#         for cl in clusters_now:
#             # qubits_ordering_clusters += list(cl)
#             qubits_ordering_all += list(cl)
#             for q in neighborhoods_tuples[cl]:
#                 if q not in qubits_ordering_all:
#                     qubits_ordering_all.append(q)

#         reversed_enum = utils.map_index_to_order(qubits_ordering_all)

#         # if tuple(clusters_now) in marginal_probs_big.keys():
#         #     probability_distribution = marginal_probs_big[tuple(clusters_now)]

#         # else:
#         noise_matrices_dictionary_local = {}
#         clusters_list_local = []
#         local_neighbors_dictionary = {}

#         for cl in clusters_now:
#             cluster_key_local = 'q' + 'q'.join([str(reversed_enum[x]) for x in cl])

#             neighbors_now = neighborhoods_tuples[cl]

#             noise_matrices_now = noise_model_description[cl]

#             # neighbors_enum = utils.map_index_to_order(neighbors_now)

#             # neighbors_local = [q for q in neighbors_now if q in qubits_ordering_all]
#             # neighbors_outside = [q for q in neighbors_now if q not in qubits_ordering_all]

#             # if len(neighbors_outside)!=0:

#             noise_matrices_dictionary_local[cluster_key_local] = noise_model_description[cl]
#             clusters_list_local.append([reversed_enum[x] for x in cl])
#             local_neighbors_dictionary[cluster_key_local] = [reversed_enum[x] for x in
#                                                              neighborhoods_tuples[cl]]

#         global_matrix_creator = GlobalNoiseMatrixCreator(
#             noise_matrices_dictionary=noise_matrices_dictionary_local,
#             clusters_list=clusters_list_local,
#             neighborhoods=local_neighbors_dictionary)

#         input_state_local = ''.join([input_state[x] for x in qubits_ordering_all])
#         marginal_now_big = []
#         full_register_local = anf.get_classical_register_bitstrings(range(len(qubits_ordering_all)))
#         for local_output_state in full_register_local:
#             marginal_now_big.append(
#                 global_matrix_creator.compute_matrix_element(input_state=input_state_local,
#                                                              output_state=local_output_state))

#         probability_distribution = np.array(marginal_now_big).reshape(len(marginal_now_big), 1)
#         marginal_probs_big[tuple(clusters_now)] = probability_distribution

#         bits_of_interest_now = [reversed_enum[q] for q in subset]

#         marginal_distro = marginals_analyzer.get_marginal_from_probability_distribution(
#             global_probability_distribution=probability_distribution,
#             bits_of_interest=bits_of_interest_now)

#         marginals[qubits_string] = marginal_distro

#         for q in subset:
#             marginals['q%s' % q] = marginals_analyzer.get_marginal_from_probability_distribution(
#                 global_probability_distribution=probability_distribution,
#                 bits_of_interest=[reversed_enum[q]])

#     return marginals





##############################################################
##################### DEPRECIATED? ###########################
##############################################################

def calculate_pairs_marginals_from_tensor_model_terrible(input_state: str,
                                                         noise_matrices_dictionary: Dict[Tuple[int],
                                                                                         np.ndarray],
                                                         needed_pairs,
                                                         get_also_1q_marginals=True
                                                         ):
    # TODO FBM: generalize to higher locality

    all_clusters = list(noise_matrices_dictionary.keys())

    pairs_marginals = {}

    for cluster, noise_matrix in noise_matrices_dictionary.items():
        pairs_in_cluster = [(qi, qj) for qi in cluster for qj in cluster if
                            qj > qi and (qi, qj) in needed_pairs]

        cluster_input_state = ''.join([input_state[x] for x in cluster])

        enumerated_qubits = utils.map_index_to_order(list(cluster))

        marginals_now = get_noisy_marginals_within_cluster(input_state=cluster_input_state,
                                                           cluster_noise_matrix=noise_matrix,
                                                           subsets_of_interest=pairs_in_cluster,
                                                           qubits_mapping=enumerated_qubits)

        pairs_marginals = {**pairs_marginals, **marginals_now}

        if get_also_1q_marginals:
            marginals_now = get_noisy_marginals_within_cluster(input_state=cluster_input_state,
                                                               cluster_noise_matrix=noise_matrix,
                                                               subsets_of_interest=[[qi] for qi in
                                                                                    cluster],
                                                               qubits_mapping=enumerated_qubits)

            pairs_marginals = {**pairs_marginals, **marginals_now}

        other_clusters = list(set(all_clusters).difference(set([cluster])))
        for cluster2 in other_clusters:
            pairs_between_clusters = [(qi, qj) for qi in cluster for qj in cluster2 if qj > qi]

            enumerated_qubits = utils.map_index_to_order(
                list(cluster) + list(cluster2))

            # print(cluster,cluster2)
            # print(enumerated_qubits)
            local_number_of_qubits = len(cluster) + len(cluster2)

            embeded_noise_matrix1 = qanf.embed_operator_in_bigger_hilbert_space(
                number_of_qubits=local_number_of_qubits,
                local_operator=noise_matrix,
                global_indices=[enumerated_qubits[x] for x in cluster]
            ).real
            embeded_noise_matrix2 = qanf.embed_operator_in_bigger_hilbert_space(
                number_of_qubits=local_number_of_qubits,
                local_operator=noise_matrices_dictionary[cluster2],
                global_indices=[enumerated_qubits[x] for x in cluster2]
            ).real

            local_tensor_noise_model = embeded_noise_matrix1 @ embeded_noise_matrix2

            input_state_now_on_two_clusters = ''.join([input_state[x] for x in
                                                       list(cluster) + list(cluster2)])

            marginals_now = get_noisy_marginals_within_cluster(
                input_state=input_state_now_on_two_clusters,
                cluster_noise_matrix=local_tensor_noise_model,
                subsets_of_interest=pairs_between_clusters,
                qubits_mapping=enumerated_qubits)

            pairs_marginals = {**pairs_marginals,
                               **marginals_now}

    return pairs_marginals




    # raise KeyError

    # # # big_noise_matrix = 1
    # # # for cl in clusters_now:
    # # #     noise_matrices_on_cluster = noise_model_description[cl]
    # # #     if len(neighborhoods_tuples[cl]) == 0:
    # # #         noise_matrix_local = noise_matrices_on_cluster['averaged']
    # # #     else:
    # # #         input_on_neighbors = ''.join(
    # # #             [input_state[x] for x in neighborhoods_tuples[cl]])
    # # #         noise_matrix_local = noise_matrices_on_cluster[input_on_neighbors]
    # # #
    # # #     big_noise_matrix = np.kron(big_noise_matrix, noise_matrix_local)
    # #
    # # input_on_cluster = ''.join([input_state[x] for x in cl])
    # #
    # # probability_distribution = big_noise_matrix[:, int(input_on_cluster, 2)]

    # really_needed_subsets = []
    # for subset in needed_subsets:
    #     qubits_string = 'q' + 'q'.join([str(x) for x in subset])
    #     bigger_subset = convert.get_qubit_indices_from_keystring(correction_indices[qubits_string])
    #     if bigger_subset not in really_needed_subsets:
    #         really_needed_subsets.append(sorted(bigger_subset))

    # marginals = {}
    # for subset in really_needed_subsets:
    #     clusters_now = []

    #     for qubit in subset:
    #         if clusters_dictionary[qubit] not in clusters_now:
    #             clusters_now.append(clusters_dictionary[qubit])

    #     qubits_with_neighbors = []
    #     for cluster in clusters_now:
    #         qubits_with_neighbors += neighborhoods[tuple(cluster)]
    #         qubits_with_neighbors += cluster

    #     qubits_with_neighbors = sorted(qubits_with_neighbors)
    #     enumerated_indices = utils.map_index_to_order(qubits_with_neighbors)
    #     clusters_enumerated = [[enumerated_indices[x] for x in cluster] for cluster in clusters_now]
    #     local_neighbors_dictionary = {}

    #     # print(neighborhoods)
    #     for cluster in clusters_now:
    #         key_qubits = 'q' + 'q'.join([str(enumerated_indices[x]) for x in cluster])
    #         local_neighbors_dictionary[key_qubits] = [enumerated_indices[x] for x in
    #                                                   neighborhoods[tuple(cluster)]]

    #     noise_matrices_dictionary_local = {'q' + 'q'.join([str(x) for x in cluster]): {} for cluster in
    #                                        clusters_enumerated}

    #     for cluster in clusters_now:
    #         key_qubits = 'q' + 'q'.join([str(x) for x in cluster])
    #         key_qubits_local = 'q' + 'q'.join([str(enumerated_indices[x]) for x in cluster])

    #         dict_now = noise_matrices_dictionary[key_qubits]

    #         noise_matrices_dictionary_local[key_qubits_local]['averaged'] = dict_now['averaged']

    #         if len(neighborhoods[tuple(cluster)]) != 0:
    #             qubits_neighbors_key = 'q' + 'q'.join([str(x) for x in neighborhoods[tuple(cluster)]])
    #             qubits_neighbors_local_key = 'q' + 'q'.join(
    #                 [str(enumerated_indices[x]) for x in neighborhoods[tuple(cluster)]])
    #             noise_matrices_dictionary_local[key_qubits_local][qubits_neighbors_local_key] = \
    #                 dict_now[qubits_neighbors_key]

    #     # print(noise_matrices_dictionary)
    #     # print(noise_matrices_dictionary_local)
    #     # raise KeyError
    #     register_size_local = len(qubits_with_neighbors)

    #     input_state_local = ''.join([input_state[x] for x in qubits_with_neighbors])

    #     full_register_local = anf.get_classical_register_bitstrings(
    #         qubit_indices=list(range(register_size_local)))

    #     global_matrix_creator = GlobalNoiseMatrixCreator(
    #         noise_matrices_dictionary=noise_matrices_dictionary_local,
    #         clusters_list=clusters_enumerated,
    #         neighborhoods=local_neighbors_dictionary)

    #     marginal_now = []
    #     # print(full_register_local)
    #     for local_output_state in full_register_local:
    #         marginal_now.append(
    #             global_matrix_creator.compute_matrix_element(input_state=input_state_local,
    #                                                          output_state=local_output_state))

    #     marginal_now = np.array(marginal_now).reshape(len(marginal_now), 1)

    #     qubits_key_final = 'q' + 'q'.join([str(x) for x in subset])
    #     marginals[qubits_key_final] = marginal_now

    # return marginals











"""class should:

    should work for any field that could appear on a class and work for numpyarrays
    to_json() function that will return dict with class __ dict __ dump
    export_json(path)
    import_json(path)
    export_pickle(path)
    import_pickle(path)
"""
from pathlib import Path
import pickle

import orjson
import numpy as np



class DataStructureBase:
    def get_dict_format(self):
        """returns this class as a json-like dictionary structure"""
        return self.__dict__

    def load_from_dict(self, dictionary):
        for key in dictionary:
            if key in self.get_dict_format():
                setattr(self, key,
                        dictionary[key])
        pass

    def to_json(self):
        """returns this class as a json structure"""
        return orjson.dumps(self, default=lambda o: o.__dict__,
                            option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_SORT_KEYS)

    def export_json(self, json_export_path, overwrite=False):
        """saves the class into json file"""
        if Path(json_export_path).is_file() and not overwrite:
            print(f"WARNING:: Omitting export to existing file: <{json_export_path}>")
        else:
            with open(json_export_path, 'wb') as outfile:
                outfile.write(self.to_json())

    def import_json(self, json_import_path):
        """imports class' fields from json file"""
        with open(json_import_path, 'rb') as outfile:
            json_dict = orjson.loads(outfile.read())
        self.load_from_dict(json_dict)

    def get_pickle(self):
        """returns this class as a pickle structure"""
        return pickle.dumps(self.get_dict_format())

    def export_pickle(self, pickle_export_path, overwrite=False):
        """saves the class into pickle file"""
        if Path(pickle_export_path).is_file() and not overwrite:
            print(f"WARNING:: Omitting export to existing file: <{pickle_export_path}>")
        else:
            with open(pickle_export_path, 'wb') as outfile:
                pickle.dump(self.get_dict_format(), outfile)

    def import_pickle(self, pickle_import_path):
        """imports class' fields from pickle file"""
        with open(pickle_import_path, 'rb') as outfile:
            pickle_dict = pickle.load(outfile)
        self.load_from_dict(pickle_dict)

class MyStructure(DataStructureBase):
    def __init__(self,data):
        super().__init__()
        self.data = data

if __name__ == "__main__":
    testcase1 = MyStructure(np.random.randint(10,size =10))
    print(testcase1.to_json())


import random
from qrem.types import DataStructureBase
from typing import Dict, Tuple, List

import numpy as np
from qrem.ctmp.model import rates_from_exp_results
import qrem.ctmp.modeltools.ground_state_estimation as gsa

class CTMPModelData(DataStructureBase):
    """The CTMP model as described in Bravyi et al.

    Parameters
    ----------
    n: int
        the number of qubits
    rates: list
        the list of error types, where each error = (i, j in, out, rate), with:
        i, j - qubits on which the error acts (i=j - one qubit error)
        in, out - input and output bits of the error
        rate - the rate of the given error
        e.g. (0, 2, "00", "11", 0.6) is the error flipping 0_0... --> 1_1...
        (1, 1, "0", "1", 0.6) is the error flipping _0_... -> _1_...
    gamma: float
        CTMP noise strength as in eq. (13); if None, it will be estimated heuristically

    """

    def __init__(self, rates: List[Tuple]=[], n: int=0, gamma: float=None):
        super().__init__()
        self.rates = rates
        self.n = n
        if gamma is None and len(rates)>0:
            self.gamma = self._estimate_gamma()
        else:
            self.gamma = gamma

    def _estimate_gamma(self) -> float:
        # temporary solution until solver issue is taken care of
        return sum([error[-1] for error in self.rates])
        # return gsa.estimate_gamma(self.n, self.rates)




# -*- coding: utf-8 -*-
"""
This file contains helper functions to work with SDPA.

Created on Fri May 16 13:52:58 2014

@author: Peter Wittek
"""
from __future__ import division, print_function
from bisect import bisect_left
import numpy as np
import os
from subprocess import call
import tempfile
from .nc_utils import convert_monomial_to_string


def parse_solution_matrix(iterator):
    solution_matrix = []
    while True:
        sol_mat = None
        in_matrix = False
        i = 0
        for row in iterator:
            if row.find('}') < 0:
                continue
            if row.startswith('}'):
                break
            if row.find('{') != row.rfind('{'):
                in_matrix = True
            numbers = row[row.rfind('{')+1:row.find('}')].strip().split(',')
            if sol_mat is None:
                sol_mat = np.empty((len(numbers), len(numbers)))
            for j, number in enumerate(numbers):
                sol_mat[i, j] = float(number)
            if row.find('}') != row.rfind('}') or not in_matrix:
                break
            i += 1
        solution_matrix.append(sol_mat)
        if row.startswith('}'):
            break
    if len(solution_matrix) > 0 and solution_matrix[-1] is None:
        solution_matrix = solution_matrix[:-1]
    return solution_matrix


def read_sdpa_out(filename, solutionmatrix=False, status=False,
                  sdp=None):
    """Helper function to parse the output file of SDPA.

    :param filename: The name of the SDPA output file.
    :type filename: str.
    :param solutionmatrix: Optional parameter for retrieving the solution.
    :type solutionmatrix: bool.
    :param status: Optional parameter for retrieving the status.
    :type status: bool.
    :param sdp: Optional parameter to add the solution to a
                          relaxation.
    :type sdp: sdp.
    :returns: tuple of two floats and optionally two lists of `numpy.array` and
              a status string
    """
    primal = None
    dual = None
    x_mat = None
    y_mat = None
    status_string = None

    with open(filename, 'r') as file_:
        for line in file_:
            if line.find("objValPrimal") > -1:
                primal = float((line.split())[2])
            if line.find("objValDual") > -1:
                dual = float((line.split())[2])
            if solutionmatrix:
                if line.find("xMat =") > -1:
                    x_mat = parse_solution_matrix(file_)
                if line.find("yMat =") > -1:
                    y_mat = parse_solution_matrix(file_)
            if line.find("phase.value") > -1:
                if line.find("pdOPT") > -1:
                    status_string = 'optimal'
                elif line.find("pFEAS") > -1:
                    status_string = 'primal feasible'
                elif line.find("pdFEAS") > -1:
                    status_string = 'primal-dual feasible'
                elif line.find("dFEAS") > -1:
                    status_string = 'dual feasible'
                elif line.find("INF") > -1:
                    status_string = 'infeasible'
                elif line.find("UNBD") > -1:
                    status_string = 'unbounded'
                else:
                    status_string = 'unknown'

    for var in [primal, dual, status_string]:
        if var is None:
            status_string = 'invalid'
            break
    if solutionmatrix:
        for var in [x_mat, y_mat]:
            if var is None:
                status_string = 'invalid'
                break

    if sdp is not None:
        sdp.primal = primal
        sdp.dual = dual
        sdp.x_mat = x_mat
        sdp.y_mat = y_mat
        sdp.status = status_string
    if solutionmatrix and status:
        return primal, dual, x_mat, y_mat, status_string
    elif solutionmatrix:
        return primal, dual, x_mat, y_mat
    elif status:
        return primal, dual, status_string
    else:
        return primal, dual


def which(program):

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def detect_sdpa(solverparameters):
    solverexecutable = "sdpa"
    if solverparameters is not None and "executable" in solverparameters:
        solverexecutable = solverparameters["executable"]
    return which(solverexecutable)


def solve_with_sdpa(sdp, solverparameters=None):
    """Helper function to write out the SDP problem to a temporary
    file, call the solver, and parse the output.

    :param sdp: The SDP relaxation to be solved.
    :type sdp: :class:`ncpol2sdpa.sdp`.
    :param solverparameters: Optional parameters to SDPA.
    :type solverparameters: dict of str.
    :returns: tuple of float and list -- the primal and dual solution of the
              SDP, respectively, and a status string.
    """
    solverexecutable = detect_sdpa(solverparameters)
    if solverexecutable is None:
        raise OSError("SDPA is not in the path or the executable provided is" +
                      " not correct")
    primal, dual = 0, 0
    tempfile_ = tempfile.NamedTemporaryFile()
    tmp_filename = tempfile_.name
    tempfile_.close()
    tmp_dats_filename = tmp_filename + ".dat-s"
    tmp_out_filename = tmp_filename + ".out"
    write_to_sdpa(sdp, tmp_dats_filename)
    command_line = [solverexecutable, "-ds", tmp_dats_filename,
                    "-o", tmp_out_filename]
    if solverparameters is not None:
        for key, value in list(solverparameters.items()):
            if key == "executable":
                continue
            elif key == "paramsfile":
                command_line.extend(["-p", value])
            else:
                raise ValueError("Unknown parameter for SDPA: " + key)
    if sdp.verbose < 1:
        with open(os.devnull, "w") as fnull:
            call(command_line, stdout=fnull, stderr=fnull)
    else:
        call(command_line)
    primal, dual, x_mat, y_mat, status = read_sdpa_out(tmp_out_filename, True,
                                                       True)
    if sdp.verbose < 2:
        os.remove(tmp_dats_filename)
        os.remove(tmp_out_filename)
    return primal+sdp.constant_term, \
        dual+sdp.constant_term, x_mat, y_mat, status


def convert_row_to_sdpa_index(block_struct, row_offsets, row):
    """Helper function to map to sparse SDPA index values.
    """
    block_index = bisect_left(row_offsets[1:], row + 1)
    width = block_struct[block_index]
    row = row - row_offsets[block_index]
    i, j = divmod(row, width)
    return block_index, i, j


def write_to_sdpa(sdp, filename):
    """Write the SDP relaxation to SDPA format.

    :param sdp: The SDP relaxation to write.
    :type sdp: :class:`ncpol2sdpa.sdp`.
    :param filename: The name of the file. It must have the suffix ".dat-s"
    :type filename: str.
    """
    # Coefficient matrices
    row_offsets = [0]
    cumulative_sum = 0
    for block_size in sdp.block_struct:
        cumulative_sum += block_size ** 2
        row_offsets.append(cumulative_sum)
    multiplier = 1
    if sdp.F.dtype == np.complex128:
        multiplier = 2
    lines = [[] for _ in range(multiplier*sdp.n_vars+1)]
    for row in range(len(sdp.F.rows)):
        if len(sdp.F.rows[row]) > 0:
            col_index = 0
            block_index, i, j = convert_row_to_sdpa_index(sdp.block_struct,
                                                          row_offsets, row)
            for k in sdp.F.rows[row]:
                value = sdp.F.data[row][col_index]
                col_index += 1
                if k == 0:
                    value *= -1
                if sdp.F.dtype == np.float64:
                    lines[k].append('{0}\t{1}\t{2}\t{3}\n'.format(
                        block_index + 1, i + 1, j + 1, value))
                else:
                    bs = sdp.block_struct[block_index]
                    if value.real != 0:
                        lines[k].append('{0}\t{1}\t{2}\t{3}\n'.format(
                            block_index + 1, i + 1, j + 1, value.real))
                        lines[k].append('{0}\t{1}\t{2}\t{3}\n'.format(
                            block_index + 1, i + bs + 1, j + bs + 1,
                            value.real))
                    if value.imag != 0:
                        lines[k + sdp.n_vars].append(
                            '{0}\t{1}\t{2}\t{3}\n'.format(
                                block_index + 1, i + 1, j + bs + 1,
                                value.imag))
                        lines[k + sdp.n_vars].append(
                            '{0}\t{1}\t{2}\t{3}\n'.format(
                                block_index + 1, j + 1, i + bs + 1,
                                -value.imag))
    file_ = open(filename, 'w')
    file_.write('"file ' + filename + ' generated by ncpol2sdpa"\n')
    file_.write(str(multiplier*sdp.n_vars) + ' = number of vars\n')
    # bloc structure
    block_struct = [multiplier*blk_size
                    for blk_size in sdp.block_struct]
    file_.write(str(len(block_struct)) + ' = number of blocs\n')
    file_.write(str(block_struct).replace('[', '(')
                .replace(']', ')'))
    file_.write(' = BlocStructure\n')
    # c vector (objective)
    objective = \
        str(list(sdp.obj_facvar)).replace('[', '').replace(']', '')
    if multiplier == 2:
        objective += ', ' + objective
    file_.write('{'+objective+'}\n')
    for k, line in enumerate(lines):
        if line == []:
            continue
        for item in line:
            file_.write('{0}\t'.format(k)+item)
    file_.close()


def convert_to_human_readable(sdp):
    """Convert the SDP relaxation to a human-readable format.

    :param sdp: The SDP relaxation to write.
    :type sdp: :class:`ncpol2sdpa.sdp`.
    :returns: tuple of the objective function in a string and a matrix of
              strings as the symbolic representation of the moment matrix
    """

    objective = ""
    indices_in_objective = []
    for i, tmp in enumerate(sdp.obj_facvar):
        candidates = [key for key, v in
                      sdp.monomial_index.items() if v == i+1]
        if len(candidates) > 0:
            monomial = convert_monomial_to_string(candidates[0])
        else:
            monomial = ""
        if tmp > 0:
            objective += "+"+str(tmp)+monomial
            indices_in_objective.append(i)
        elif tmp < 0:
            objective += str(tmp)+monomial
            indices_in_objective.append(i)

    matrix_size = 0
    cumulative_sum = 0
    row_offsets = [0]
    block_offset = [0]
    for bs in sdp.block_struct:
        matrix_size += abs(bs)
        cumulative_sum += bs ** 2
        row_offsets.append(cumulative_sum)
        block_offset.append(matrix_size)

    matrix = []
    for i in range(matrix_size):
        matrix_line = ["0"] * matrix_size
        matrix.append(matrix_line)

    for row in range(len(sdp.F.rows)):
        if len(sdp.F.rows[row]) > 0:
            col_index = 0
            for k in sdp.F.rows[row]:
                value = sdp.F.data[row][col_index]
                col_index += 1
                block_index, i, j = convert_row_to_sdpa_index(
                    sdp.block_struct, row_offsets, row)
                candidates = [key for key, v in
                              sdp.monomial_index.items()
                              if v == k]
                if len(candidates) > 0:
                    monomial = convert_monomial_to_string(candidates[0])
                else:
                    monomial = ""
                offset = block_offset[block_index]
                if matrix[offset+i][offset+j] == "0":
                    matrix[offset+i][offset+j] = ("%s%s" % (value, monomial))
                else:
                    if value.real > 0:
                        matrix[offset+i][offset+j] += ("+%s%s" % (value,
                                                                  monomial))
                    else:
                        matrix[offset+i][offset+j] += ("%s%s" % (value,
                                                                 monomial))
    return objective, matrix


def write_to_human_readable(sdp, filename):
    """Write the SDP relaxation to a human-readable format.

    :param sdp: The SDP relaxation to write.
    :type sdp: :class:`ncpol2sdpa.sdp`.
    :param filename: The name of the file.
    :type filename: str.
    """
    objective, matrix = convert_to_human_readable(sdp)
    f = open(filename, 'w')
    f.write("Objective:" + objective + "\n")
    for matrix_line in matrix:
        f.write(str(list(matrix_line)).replace('[', '').replace(']', '')
                .replace('\'', ''))
        f.write('\n')
    f.close()



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""

import copy
from typing import Optional, List
from collections import Counter

import numpy as np
from tqdm import tqdm

from qrem.functions_qrem import functions_data_analysis as fda

from pyquil.gates import RX, RZ, MEASURE
from pyquil.quil import Program
from pyquil.quilatom import MemoryReference

from qrem.common.printer import qprint


#MOVE_TO qrem.common.math ? 
# simpler circuits
__dictionary_with_angles_pauli_eigenstates__ = \
    {"Z+": {"RZ": (0, 0, 0),
            "RX": (0, 0)},

     "Z-": {"RZ": (-0.9199372448290238, np.pi, 2.2216554087607694),
            "RX": (np.pi / 2, -np.pi / 2)},

     "X+": {"RZ": (np.pi / 2, np.pi / 2, 0),
            "RX": (np.pi / 2, 0)},

     "X-": {"RZ": (np.pi / 2, -np.pi / 2, 0),
            "RX": (np.pi / 2, 0)},

     "Y+": {"RZ": (np.pi / 2, np.pi, 0),
            "RX": (np.pi / 2, 0)},

     "Y-": {"RZ": (np.pi / 2, 0, 0),
            "RX": (np.pi / 2, 0)},
     }

enumerated_labels_pauli_eigenstates = dict(enumerate(['Z+', 'Z-',
                                                      "X+", 'X-',
                                                      "Y+", "Y-"]))


def _apply_elementary_gate_native(quantum_program,
                                  qubit_index: int,
                                  gate_label: str):
    # This funcion decomposes elementary quantum gates (I, X, Y, Z, H, S, T) into gates native to rigetti devices
    if gate_label.upper() == 'I':
        # quantum_circuit+=Instruction(Gate.I(), qubit_index)
        quantum_program = quantum_program.rz(qubit_index, 0)

    elif gate_label.upper() == 'X':
        # X
        quantum_program = quantum_program.rz(qubit_index, np.pi)
        quantum_program = quantum_program.rx(qubit_index, np.pi / 2)
        quantum_program = quantum_program.rz(qubit_index, np.pi)
        quantum_program = quantum_program.rx(qubit_index, -np.pi / 2)

    elif gate_label.upper() == 'Y':
        # Y
        quantum_program = quantum_program.rz(qubit_index, -2.446121240088584)
        quantum_program = quantum_program.rx(qubit_index, np.pi / 2)
        quantum_program = quantum_program.rz(qubit_index, np.pi)
        quantum_program = quantum_program.rx(qubit_index, -np.pi / 2)
        quantum_program = quantum_program.rz(qubit_index, -2.446121240088584)

    elif gate_label.upper() == 'Z':
        # Z
        quantum_program = quantum_program.rz(qubit_index, np.pi)


    elif gate_label.upper() == 'H':
        # H
        quantum_program = quantum_program.rz(qubit_index, np.pi)
        quantum_program = quantum_program.rx(qubit_index, np.pi / 2)
        quantum_program = quantum_program.rz(qubit_index, np.pi / 2)
        quantum_program = quantum_program.rx(qubit_index, -np.pi / 2)


    elif gate_label.upper() == 'S':
        # S
        quantum_program = quantum_program.rz(qubit_index, np.pi / 2)


    elif gate_label.upper() == 'S*':
        # Sdag
        quantum_program = quantum_program.rz(qubit_index, -np.pi / 2)

    elif gate_label.upper() == 'T':
        # T
        quantum_program = quantum_program.rz(qubit_index, np.pi / 4)

    elif gate_label.upper() == 'T*':
        # Tdag
        quantum_program = quantum_program.rz(qubit_index, -np.pi / 4)

    else:
        raise ValueError(f"Gate label: '{gate_label}' not recognized.")

    return quantum_program


def _declare_generic_DOT_gate_parametric(quantum_program,
                                         qubit_index: int,
                                         number_of_rz_rotations=3,
                                         number_of_rx_rotations=2,
                                         name_suffix='',
                                         name_prefix=''):
    quantum_program.declare(f"{name_prefix}RZ_q-{qubit_index}{name_suffix}", "REAL",
                            number_of_rz_rotations)
    quantum_program.declare(f"{name_prefix}RX_q-{qubit_index}{name_suffix}", "REAL",
                            number_of_rx_rotations)

    return quantum_program


def _apply_generic_DOT_gate_parametric(quantum_program,
                                       qubit_index_physical: int):
    """

    :param quantum_program:
    :type quantum_program:
    :param qubit_index_physical:
    :type qubit_index_physical:

    :return:
    :rtype:
    """

    quantum_program += RZ(MemoryReference(f"RZ_q-{qubit_index_physical}", 0), qubit_index_physical)
    quantum_program += RX(MemoryReference(f"RX_q-{qubit_index_physical}", 0), qubit_index_physical)
    quantum_program += RZ(MemoryReference(f"RZ_q-{qubit_index_physical}", 1), qubit_index_physical)
    quantum_program += RX(MemoryReference(f"RX_q-{qubit_index_physical}", 1), qubit_index_physical)
    quantum_program += RZ(MemoryReference(f"RZ_q-{qubit_index_physical}", 2), qubit_index_physical)

    return quantum_program


def create_memory_map_DOT(list_of_keys: List[str],
                          qubit_indices: List[int]):
    if isinstance(list_of_keys, dict):
        list_of_keys = list(list_of_keys.keys())
    elif isinstance(list_of_keys, str):
        list_of_keys = [list_of_keys]

    number_of_qubits = len(qubit_indices)

    memory_map = {}
    for qubit_index_enumerated in range(number_of_qubits):
        qubit_index_physical = qubit_indices[qubit_index_enumerated]
        memory_map[f"RZ_q-{qubit_index_physical}"] = []
        memory_map[f"RX_q-{qubit_index_physical}"] = []

    for circuit_key in list_of_keys:
        for qubit_index_enumerated in range(number_of_qubits):
            gate_index = int(circuit_key[qubit_index_enumerated])
            angles_now = __dictionary_with_angles_pauli_eigenstates__[
                enumerated_labels_pauli_eigenstates[gate_index]]
            angles_RZ_now = list(angles_now['RZ'])
            angles_RX_now = list(angles_now['RX'])

            qubit_index_physical = qubit_indices[qubit_index_enumerated]
            memory_map[f"RZ_q-{qubit_index_physical}"].append(angles_RZ_now)
            memory_map[f"RX_q-{qubit_index_physical}"].append(angles_RX_now)

    return memory_map


def get_generic_base_program_DOT(qubit_indices: List[int],
                                 compilation_name: Optional[str] = 'parametric-native'):
    # if sdk_name.upper() in ['PYQUIL-FOR-AZURE-QUANTUM']:

    classical_register_size = len(qubit_indices)

    quantum_program = Program()
    quantum_program.declare("ro",
                            "BIT",
                            classical_register_size)

    if compilation_name.upper() in ['PARAMETRIC-NATIVE', 'NATIVE-PARAMETRIC']:
        for qubit_index in qubit_indices:
            quantum_program = _declare_generic_DOT_gate_parametric(quantum_program=quantum_program,
                                                                   qubit_index=qubit_index)

        for qubit_index in qubit_indices:
            quantum_program = _apply_generic_DOT_gate_parametric(quantum_program=quantum_program,
                                                                 qubit_index_physical=qubit_index)
    else:
        raise ValueError(f"Wrong compilation method: '{compilation_name}'")

    for qubit_index_enumerated in range(len(qubit_indices)):
        quantum_program += MEASURE(qubit_indices[qubit_index_enumerated],
                                   MemoryReference(f"ro",
                                                   qubit_index_enumerated))

    return quantum_program


def get_backend_wrapper(backend_name,
                        sdk_name):
    if sdk_name.upper() in ['PYQUIL-FOR-AZURE-QUANTUM']:

        from pyquil_for_azure_quantum import get_qpu, get_qvm, get_qc

        if backend_name.upper() in ['ASPEN-M-2', 'ASPEN-11']:
            backend_instance = get_qpu(backend_name)
        else:

            if backend_name.upper() in ['QVM']:
                backend_instance = get_qvm()

            else:
                if backend_name.upper() in ['ASPEN-M-2-QVM', 'ASPEN-11-QVM']:
                    # test = backend_name
                    # print('test', backend_name[0:-4])
                    backend_instance = get_qc(backend_name[0:-4],
                                              as_qvm=True,
                                              compiler_timeout=60 * 30,
                                              execution_timeout=60 * 60 * 24)
                else:
                    backend_instance = get_qc(backend_name,
                                              as_qvm=True,
                                              compiler_timeout=60 * 30,
                                              execution_timeout=60 * 60 * 24
                                              )

    elif sdk_name.upper() in ['PYQUIL']:
        from pyquil import get_qc
        backend_instance = get_qc(backend_name)
    else:
        raise ValueError(f"Wrong backend name: '{backend_name}'.")

    return backend_instance


def run_batches_parametric(
        backend_name: str,
        sdk_name: str,
        number_of_shots: int,
        qubit_indices: List[int],
        base_program,
        memory_map,
        compilation_method: Optional[str] = 'parametric-native',
):
    # TODO FBM: add batching jobs

    if base_program is None:
        base_program = get_generic_base_program_DOT(qubit_indices=qubit_indices)

    base_program.wrap_in_numshots_loop(shots=number_of_shots)

    backend_instance = get_backend_wrapper(backend_name=backend_name,
                                           sdk_name=sdk_name)

    target_number_of_circuits = len(list(memory_map.values())[0])

    qprint('\nSending jobs to execution on: ', backend_name + '.')
    qprint('Number of shots: ', str(number_of_shots) + ' .')
    qprint('Target number of circuits: ', str(target_number_of_circuits) + ' .')
    qprint("Compilation method:", compilation_method)

    # TODO FBM: add handling errors, waiting etc

    if compilation_method.upper() in ['NATIVE-PARAMETRIC', 'PARAMETRIC-NATIVE']:

        # get native quil commands
        native_quil = backend_instance.compiler.quil_to_native_quil(base_program,
                                                                    # compilatio=60*60
                                                                    )
        if backend_name.upper() in ['ASPEN-M-2', 'ASPEN-11']:

            # create executable
            executable = backend_instance.compile(native_quil,
                                                  to_native_gates=False)
            results = backend_instance.run_batch(executable,
                                                 memory_map)
        else:
            results = []
            for circuit_index in tqdm(range(target_number_of_circuits)):
                native_quil_now = copy.deepcopy(native_quil)
                for angle_label, angles_values_list in memory_map.items():
                    angles_values_now = angles_values_list[circuit_index]
                    for index_angle in range(len(angles_values_now)):
                        native_quil_now.write_memory(region_name=angle_label,
                                                     value=angles_values_now[index_angle],
                                                     offset=index_angle)

                executable = backend_instance.compile(program=native_quil_now,
                                                      to_native_gates=True)
                results_now = backend_instance.run(executable)
                results.append(results_now)
            # executable = backend_instance.compile(native_quil,
            #                                       to_native_gates=True)
            # results = backend_instance.run_experiment(base_program,
            #                                      memory_map=memory_map)



    else:
        raise ValueError(f"Wrong compilation method: '{compilation_method}'.")

    return results


def convert_results_to_counts(single_result):
    bitstrings_list = single_result.readout_data['ro']
    to_strings = [''.join([str(s) for s in x]) for x in bitstrings_list]

    return dict(Counter(to_strings))


def _convert_results_to_counts_tuples(single_result):
    bitstrings_list = single_result.readout_data['ro']
    to_strings = [tuple(x) for x in bitstrings_list]

    return dict(Counter(to_strings))


def convert_results_to_counts_dictionaries_DOT(list_of_circuits_labels,
                                               results_list):
    results_dictionary = {}

    for circuit_index in tqdm(range(len(list_of_circuits_labels))):
        label_now = ''.join([str(s) for s in list_of_circuits_labels[circuit_index]])
        result_now = results_list[circuit_index]
        counts_now = _convert_results_to_counts_tuples(result_now)

        if label_now not in results_dictionary.keys():
            new_dict = counts_now
        else:
            existing_counts = results_dictionary[label_now]
            new_dict = fda.merge_multiple_counts_dictionaries([existing_counts, counts_now])

            # x, y = counts_now, existing_counts
            # new_dict = {k: x.get(k, 0) + y.get(k, 0) for k in set(x) | set(y)}
            #

        results_dictionary[label_now] = new_dict

    formatted_results_dictionary = {
        label_circuit: {''.join([str(x) for x in bitstring_tuple]): ticks for bitstring_tuple, ticks in
                        counts_dictionary.items()} for label_circuit, counts_dictionary in
        tqdm(results_dictionary.items())}

    return formatted_results_dictionary
#
# def run_batches(batches,
#                 backend_name,
#                 number_of_shots,
#                 sdk_name='pyquil',
#                 saving_IDs_dictionary={'saving': False,
#                                        'directory': None,
#                                        'file_name': None,
#                                        'dictionary_to_save': {}},
#                 provider_data: Optional[Dict[str, str]] = None):
#     qprint('\nSending jobs to execution on: ', backend_name + '.')
#     qprint('Number of shots: ', str(number_of_shots) + ' .')
#     qprint('Target number of jobs: ', str(len(batches)) + ' .')
#
#     batches_wrapped_with_shots = \
#         [[program.wrap_in_numshots_loop(number_of_shots) for program in batch] for batch in batches]
#
#     if backend_name in ['qvm']:
#         if sdk_name.lower() in ['pyquil-for-azure-quantum']:
#             pass
#
#
#
#         elif sdk_name.lower() in ['pyquil']:
#             # TODO FBM: finish this
#             raise ValueError("pyquil not added yet")
#
#
#     else:
#         if sdk_name.lower() in ['pyquil-for-azure-quantum']:
#             pass
#
#
#
#         elif sdk_name.lower() in ['pyquil']:
#             # TODO FBM: finish this
#             raise ValueError("pyquil not added yet")



import abc
from typing import List


class MarginalsAnalyzerInterface(abc.ABC):
    """
    This is interface for classes that will analyze marginal probability distributions.
    It requires those child classes to have basic functionalities that should be included.
    for analyzing marginals_dictionary.
    """

    @property
    @abc.abstractmethod
    def results_dictionary(self) -> dict:
        # dictionary of experimental results
        raise NotImplementedError

    @results_dictionary.setter
    @abc.abstractmethod
    def results_dictionary(self, results_dictionary: dict) -> None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def marginals_dictionary(self) -> dict:
        # Dictionary storing marginal probability distributions
        raise NotImplementedError

    @marginals_dictionary.setter
    @abc.abstractmethod
    def marginals_dictionary(self, marginals_dictionary: dict) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def results_dictionary_update(self, results_dictionary_new: dict) -> None:
        # Updating dictionary with results
        raise NotImplementedError

    @abc.abstractmethod
    def marginals_dictionary_update(self, marginals_dictionary_new: dict) -> None:
        # Updating dictionary with new marginals_dictionary
        raise NotImplementedError

    @abc.abstractmethod
    def compute_unnormalized_marginals(self,
                                       experiment_key: str,
                                       subsets: List[List[int]]) -> dict:
        """Computes marginals_dictionary for input subsets_list of qubits"""
        raise NotImplementedError



import numpy as np
import math
import qrem.common.constants as constants
import qrem.common.experiment.tomography as tomography
import itertools
from typing import List, Dict, Optional, Callable, Tuple

#np.full(N,value,dtype=consts.CIRCUIT_DATA_TYPE)




def generate_combinatorial_circuits(experiment_type: str,
                               number_of_qubits: int,
                               subset_locality: int, symbols:Optional[List] = None)-> np.array:
    try:
        number_of_symbols = constants.EXPERIMENT_TYPE_SYMBLOS[experiment_type.lower()]
    except:
        print("Error: wrong experiment type")
    if symbols==None:
        symbols = np.array(range(number_of_symbols), dtype = constants.CIRCUIT_DATA_TYPE)
    s = len(symbols)
    n = number_of_qubits
    k = subset_locality

    circuits = [[symbol for i in range(n)] for symbol in symbols]


    combinations_list = list(itertools.product(symbols, repeat=k))
    for c in combinations_list:
        if c == tuple(c[0] for i in range(len(c))):
            combinations_list.remove(c)
    #circuits.append(new_entry)
    m = np.ceil(np.log(n)/np.log(k))
    N = k**m

    for i in np.arange(m):
        for c in combinations_list:
            circ = [c[(int(j/k**i))%k] for j in np.arange(N)]
            circuits.append(circ[0:n])

    if k>2:
        pass

    print(len(circuits))
    return np.array(circuits,dtype = constants.CIRCUIT_DATA_TYPE)

def complete_on_subset(experiment_type: str, circuit_list: List, subset: Tuple, absent_symbols: List):

    try:
        number_of_symbols = constants.EXPERIMENT_TYPE_SYMBLOS[experiment_type.lower()]
    except:
        print("Error: wrong experiment type")
    symbols_list = list(range(number_of_symbols))
    number_of_qubits = len(circuit_list[0])
    new_circuit_list = circuit_list
    locality = len(subset)

    for a in absent_symbols:
        new_circuit =np.array( [[np.random.randint(0, number_of_symbols) for i in range(number_of_qubits)]],dtype = constants.CIRCUIT_DATA_TYPE)
        for i in range(locality):
            new_circuit[0][subset[i]] = a[i]
        new_circuit_list = np.append(new_circuit_list,new_circuit,axis=0)

    return new_circuit_list

def find_partition(subset_list, n):
    subset = subset_list[0]
    k = len(subset)
    p = int(n/k)
    partition = [subset]
    partition_set = set(subset)
    for s in subset_list:
        if partition_set.isdisjoint(set(s)):
            partition_set = partition_set.union(set(s))
            partition.append(s)
        if len(partition)==p:
            break

    return partition





if __name__ == "__main__":
    experiments = ["DDOT","QDOT"]
    number_of_qubits = [128]
    locality = [3]
    for e in experiments:
        for k in locality:
            for n in number_of_qubits:
                print(e,n,k)
                symbols = np.arange(constants.EXPERIMENT_TYPE_SYMBLOS[e.lower()])
                #print(number_of_circuits(e,n,k))
                circuits = generate_combinatorial_circuits(e,n,k)


                incomplete_dict = tomography.check_completeness(e,circuits,k)
                nk = math.factorial(n)/(math.factorial(k)*math.factorial(n-k))
                print(len(incomplete_dict),"incomplete subsets out of ",nk,"fraction: ",1.*len(incomplete_dict)/nk)
                print(incomplete_dict.keys())

                print(find_partition(list(incomplete_dict.keys()),n))


                """
                if k==3:
                    while incomplete_dict != {}:
                        print(len(incomplete_dict.keys()))
                        circuit_list_update = tomography.complete_on_subset(e, circuits,
                                                                 list(incomplete_dict.keys())[0],
                                                                 list(incomplete_dict.values())[0])

                        dict_update = tomography.check_completeness(e, circuit_list_update, k)
                        circuits = circuit_list_update
                        incomplete_dict = dict_update
"""



"""4. napisać funkcję run_simulation_experiment.py, która (1) stworzy przykladowy model(e), (2) zwróci nam jakąś przykladową symulację."""

from qrem.common.printer import qprint, errprint, warprint



if __name__ == "__main__":
    warprint("=============================")
    warprint("START: Executing SIMULATION Experiment pipeline")
    warprint("=============================")